import copy
import inspect
import math
from collections import OrderedDict

import numpy as np
import torch
import torch.nn as nn
from torch.nn import functional as F

from agilerl.networks.evolvable_mlp import EvolvableMLP


class EvolvableGPT(nn.Module):
    """The Evolvable GPT class.

    :param n_layer: Number of transformer block layers, defaults to 12
    :type encoder_layers: int, optional
    :param vocab_size: Vocabulary size, defaults to 50257
    :type vocab_size: int, optional
    :param n_embd: Transformer embedding dimension size, defaults to 768
    :type n_embd: int, optional
    :param n_head: Number of heads in the multiheadattention models, defaults to 12
    :type n_head: int, optional
    :param dim_feedfwd: Size of transformer block hidden layer, defaults to 3072 (4*768)
    :type dim_feedfwd: int, optional
    :param block_size: Transformer block context size, defaults to 1024
    :type block_size: int, optional
    :param dropout: Dropout value, defaults to 0.0
    :type dropout: float, optional
    :param activation: Activation function of transformer intermediate layer, defaults to 'GELU'
    :type activation: str, optional
    :param layer_norm_eps: Epsilon value in layer normalization components, defaults to 1e-5
    :type layer_norm_eps: float, optional
    :param min_layers: Minimum number of transformer block layers, defaults to 8
    :type min_layers: int, optional
    :param max_layers: Maximum number of transformer block layers, defaults to 16
    :type max_layers: int, optional
    :param bias: Use bias in Linears and LayerNorms, defaults to True
    :type bias: bool, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    """

    def __init__(
        self,
        n_layer: int = 12,
        vocab_size: int = 50257,
        n_embd: int = 768,
        n_head: int = 12,
        dim_feedfwd: int = 3072,
        block_size: int = 1024,
        dropout: float = 0.0,
        activation: str = "GELU",
        layer_norm_eps: float = 1e-5,
        min_layers: int = 8,
        max_layers: int = 16,
        bias: bool = True,
        device="cpu",
        accelerator=None,
    ):
        assert isinstance(n_layer, int), "Number of layers must be an integer."
        assert n_layer >= 1, "Number of layers must be greater than or equal to one."
        assert isinstance(vocab_size, int), "Vocabulary size must be an integer."
        assert vocab_size >= 1, "Vocabulary size must be greater than or equal to one."
        assert isinstance(n_embd, int), "Embedding dimension must be an integer."
        assert n_embd >= 1, "Embedding dimension must be greater than or equal to one."
        assert isinstance(n_head, int), "Number of attention heads must be an integer."
        assert (
            n_head >= 1
        ), "Number of attention heads must be greater than or equal to one."
        assert isinstance(
            dim_feedfwd, int
        ), "Feed forward dimension must be an integer."
        assert (
            dim_feedfwd >= 1
        ), "Feed forward dimension must be greater than or equal to one."
        assert isinstance(block_size, int), "Block size must be an integer."
        assert block_size >= 1, "Block size must be greater than or equal to one."
        assert isinstance(dropout, (float, int)), "Dropout must be a float."
        assert 0 <= dropout <= 1, "Dropout must be between zero and one (inclusive)."
        assert isinstance(layer_norm_eps, float), "Layer norm epsilon must be a float."
        assert layer_norm_eps > 0, "Layer norm epsilon must be greater than zero."
        assert isinstance(
            min_layers, int
        ), "Minimum number of layers must be an integer."
        assert (
            min_layers >= 1
        ), "Minimum number of layers must be greater than or equal to one."
        assert isinstance(
            max_layers, int
        ), "Maximum number of layers must be an integer."
        assert (
            max_layers >= 1
        ), "Maximum number of layers must be greater than or equal to one."
        assert (
            max_layers >= min_layers
        ), "Maximum number of layers must be greater than or equal to minimum number of layers."
        assert isinstance(bias, bool), "Bias flag must be boolean value True or False."

        super().__init__()

        self.arch = "gpt"
        self.n_layer = n_layer
        self.vocab_size = vocab_size
        self.n_embd = n_embd
        self.n_head = n_head
        self.dim_feedfwd = dim_feedfwd
        self.block_size = block_size
        self.dropout = dropout
        self.activation = activation
        self.layer_norm_eps = layer_norm_eps
        self.min_layers = min_layers
        self.max_layers = max_layers
        self.bias = bias
        self.device = device
        self.accelerator = accelerator

        self.transformer = self.create_net()
        self.transformer = self.transformer.to(self.device)
        self.transformer_keys = list(self.transformer.keys())

        self.lm_head = nn.Linear(self.n_embd, self.vocab_size, bias=False)

        self.transformer.wte.weight = self.lm_head.weight

        # init all weights
        self.apply(self._init_weights)
        # apply special scaled init to the residual projections, per GPT-2 paper
        for pn, p in self.named_parameters():
            if pn.endswith("c_proj.weight"):
                torch.nn.init.normal_(
                    p, mean=0.0, std=0.02 / math.sqrt(2 * self.n_layer)
                )

        # report number of parameters
        # print("number of parameters: %.2fM" % (self.get_num_params() / 1e6,))

    def get_num_params(self, non_embedding=True):
        """Return the number of parameters in the model.
        For non-embedding count (default), the position embeddings get subtracted.
        The token embeddings would too, except due to the parameter sharing these
        params are actually used as weights in the final layer, so we include them.
        """
        n_params = sum(p.numel() for p in self.parameters())
        if non_embedding:
            n_params -= self.transformer.wpe.weight.numel()
        return n_params

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

    def get_activation(self, activation_names):
        """Returns activation function for corresponding activation name.

        :param activation_names: Activation function name
        :type activation_names: str
        """
        activation_functions = {
            "Tanh": nn.Tanh,
            "Identity": nn.Identity,
            "ReLU": nn.ReLU,
            "ELU": nn.ELU,
            "Softsign": nn.Softsign,
            "Sigmoid": nn.Sigmoid,
            "Softplus": nn.Softplus,
            "Softmax": nn.Softmax,
            "LeakyReLU": nn.LeakyReLU,
            "PReLU": nn.PReLU,
            "GELU": new_gelu,
        }

        return activation_functions[activation_names]()

    def create_net(self):
        """Creates and returns transformer neural network."""
        net_dict = OrderedDict()
        net_dict["wte"] = nn.Embedding(self.vocab_size, self.n_embd)
        net_dict["wpe"] = nn.Embedding(self.block_size, self.n_embd)
        net_dict["drop"] = nn.Dropout(self.dropout)
        net_dict["h"] = nn.ModuleList(
            [
                Block(
                    self.n_embd,
                    self.n_head,
                    self.bias,
                    self.dropout,
                    self.block_size,
                    self.dim_feedfwd,
                    self.activation,
                    self.layer_norm_eps,
                )
                for _ in range(self.n_layer)
            ]
        )
        net_dict["ln_f"] = LayerNorm(self.n_embd, bias=self.bias)
        return nn.ModuleDict(net_dict)

    def forward(
        self,
        idx=None,
        tok_emb=None,
        targets=None,
        attn_mask=None,
        past_key_values=None,
        pos=None,
        is_causal=True,
    ):
        """Forward pass through evolvable GPT model.

        :param idxs: Input ids
        :type idxs: torch.Tensor
        :param targets: Target ids
        :type targets: torch.Tensor
        """
        if idx is not None:
            device = idx.device
            t = idx.size(1)
        else:
            device = tok_emb.device
            t = tok_emb.size(-2)

        assert (
            t <= self.block_size
        ), f"Cannot forward sequence of length {t}, block size is only {self.block_size}"

        presents = ()
        all_hidden_states = ()
        if past_key_values is None:
            past_length = 0
            past_key_values = tuple([None] * self.n_layer)
        else:
            past_length = past_key_values[0][0].size(-2)

        if pos is not None:
            pos = pos.view(-1, t)
        else:
            pos = torch.arange(
                past_length, t + past_length, dtype=torch.long, device=device
            ).unsqueeze(
                0
            )  # shape (1, t)

        # forward the GPT model itself
        # token embeddings of shape (b, t, n_embd)
        if tok_emb is None:
            tok_emb = self.transformer.wte(idx)
        # position embeddings of shape (1, t, n_embd)
        pos_emb = self.transformer.wpe(pos)
        x = self.transformer.drop(tok_emb + pos_emb)
        all_hidden_states = all_hidden_states + (x,)
        for block, layer_past in zip(self.transformer.h, past_key_values):
            # torch.cuda.set_device(x.device)
            # Ensure layer_past is on same device as hidden_states (might not be correct)
            if layer_past is not None:
                layer_past = tuple(past_state.to(x.device) for past_state in layer_past)
            x, pres = block(x, attn_mask, layer_past, is_causal)
            all_hidden_states = all_hidden_states + (x,)
            presents = presents + (pres,)
        x = self.transformer.ln_f(x)
        all_hidden_states = all_hidden_states + (x,)

        logits = self.lm_head(x)
        if targets is not None:
            # if we are given some desired targets also calculate the loss
            # logits = self.lm_head(x)
            loss = F.cross_entropy(
                logits.view(-1, logits.size(-1)),
                targets.view(-1).type(torch.LongTensor).to(self.device),
                ignore_index=-1,
            )
        else:
            # inference-time mini-optimization: only forward the lm_head on the very last
            # position
            # note: using list [-1] to preserve the time dim
            # logits = self.lm_head(x[:, [-1], :])
            loss = None

        return logits, all_hidden_states, presents, loss

    def crop_block_size(self, block_size):
        # model surgery to decrease the block size if necessary
        # e.g. we may load the GPT2 pretrained model checkpoint (block size 1024)
        # but want to use a smaller block size for some smaller, simpler model
        assert block_size <= self.block_size
        self.block_size = block_size
        self.transformer.wpe.weight = nn.Parameter(
            self.transformer.wpe.weight[:block_size]
        )
        for block in self.transformer.h:
            if hasattr(block.attn, "attention_bias"):
                block.attn.attention_bias = block.attn.attention_bias[
                    :, :, :block_size, :block_size
                ]

    @classmethod
    def from_pretrained(cls, model_type, override_args=None, custom_sd=None):
        assert model_type in {"gpt2", "gpt2-medium", "gpt2-large", "gpt2-xl"}
        override_args = override_args or {}  # default to empty dict
        from transformers import GPT2Config, GPT2LMHeadModel

        print("loading weights from pretrained gpt: %s" % model_type)
        config_args = {
            "gpt2": dict(n_layer=12, n_head=12, n_embd=768),  # 124M params
            "gpt2-medium": dict(n_layer=24, n_head=16, n_embd=1024),  # 350M params
            "gpt2-large": dict(n_layer=36, n_head=20, n_embd=1280),  # 774M params
            "gpt2-xl": dict(n_layer=48, n_head=25, n_embd=1600),  # 1558M params
        }[model_type]
        config_args["vocab_size"] = 50257  # always 50257 for GPT model checkpoints
        config_args["block_size"] = 1024  # always 1024 for GPT model checkpoints
        config_args["bias"] = True  # always True for GPT model checkpoints
        # we can override the dropout rate, if desired
        if "dropout" in override_args:
            print(f"overriding dropout rate to {override_args['dropout']}")
            config_args["dropout"] = override_args["dropout"]

        if custom_sd is not None:
            if "vocab_size" in override_args:
                print(f"overriding vocab_size to {override_args['vocab_size']}")
                config_args["vocab_size"] = override_args["vocab_size"]
            model = EvolvableGPT(**config_args)
            sd_hf = torch.load(custom_sd)
            config = GPT2Config(**config_args)
            model_hf = GPT2LMHeadModel(config)
            sd_hf = {k.split("model.")[-1]: v for k, v in sd_hf.items()}
            model_hf.load_state_dict(sd_hf)
        else:
            # create a from-scratch initialized Evolvable GPT model
            model = EvolvableGPT(**config_args)
            # init a huggingface/transformers model
            model_hf = GPT2LMHeadModel.from_pretrained(model_type)
            sd_hf = model_hf.state_dict()

        sd = model.state_dict()
        sd_keys = sd.keys()
        # discard this mask / buffer, not a param
        sd_keys = [k for k in sd_keys if not k.endswith(".attn.attention_bias")]

        # copy while ensuring all of the parameters are aligned and match in names
        # and shapes
        sd_keys_hf = sd_hf.keys()
        sd_keys_hf = [
            k for k in sd_keys_hf if not k.endswith(".attn.masked_bias")
        ]  # ignore these, just a buffer
        sd_keys_hf = [
            k for k in sd_keys_hf if not k.endswith(".attn.bias")
        ]  # same, just the mask (buffer)
        transposed = [
            "attn.c_attn.weight",
            "attn.c_proj.weight",
            "mlp.c_fc.weight",
            "mlp.c_proj.weight",
        ]
        # basically the openai checkpoints use a "Conv1D" module, but we only want to use a vanilla Linear
        # this means that we have to transpose these weights when we import them
        assert len(sd_keys_hf) == len(
            sd_keys
        ), f"mismatched keys: {len(sd_keys_hf)} != {len(sd_keys)}"
        for khf, k in zip(sd_keys_hf, sd_keys):
            if any(khf.endswith(w) for w in transposed):
                # special treatment for the Conv1D weights we need to transpose
                assert sd_hf[khf].shape[::-1] == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[khf].t())
            else:
                # vanilla copy over the other parameters
                assert sd_hf[khf].shape == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[khf])
        return model

    def configure_optimizers(self, weight_decay, learning_rate, betas, device_type):
        """
        This long function is unfortunately doing something very simple and is being very defensive:
        We are separating out all parameters of the model into two buckets: those that will experience
        weight decay for regularization and those that won't (biases, and layernorm/embedding weights).
        We are then returning the PyTorch optimizer object.
        """

        # separate out all parameters to those that will and won't experience
        # regularizing weight decay
        decay = set()
        no_decay = set()
        whitelist_weight_modules = (torch.nn.Linear,)
        blacklist_weight_modules = (torch.nn.LayerNorm, LayerNorm, torch.nn.Embedding)
        for mn, m in self.named_modules():
            for pn, p in m.named_parameters():
                fpn = f"{mn}.{pn}" if mn else pn  # full param name
                # random note: because named_modules and named_parameters are recursive
                # we will see the same tensors p many many times. but doing it this way
                # allows us to know which parent module any tensor p belongs to...
                if pn.endswith("bias"):
                    # all biases will not be decayed
                    no_decay.add(fpn)
                elif pn.endswith("weight") and isinstance(m, whitelist_weight_modules):
                    # weights of whitelist modules will be weight decayed
                    decay.add(fpn)
                elif pn.endswith("weight") and isinstance(m, blacklist_weight_modules):
                    # weights of blacklist modules will NOT be weight decayed
                    no_decay.add(fpn)

        # subtle: 'transformer.wte.weight' and 'lm_head.weight' are tied, so they
        # will appear in the no_decay and decay sets respectively after the above.
        # In addition, because named_parameters() doesn't return duplicates, it
        # will only return the first occurrence, key'd by 'transformer.wte.weight', below.
        # so let's manually remove 'lm_head.weight' from decay set. This will include
        # this tensor into optimization via transformer.wte.weight only, and not
        # decayed.
        decay.remove("lm_head.weight")

        # validate that we considered every parameter
        param_dict = {pn: p for pn, p in self.named_parameters()}
        inter_params = decay & no_decay
        union_params = decay | no_decay
        assert (
            len(inter_params) == 0
        ), "parameters {} made it into both decay/no_decay sets!".format(
            str(inter_params)
        )
        assert (
            len(param_dict.keys() - union_params) == 0
        ), "parameters {} were not separated into either decay/no_decay set!".format(
            str(param_dict.keys() - union_params),
        )

        # create the pytorch optimizer object
        optim_groups = [
            {
                "params": [param_dict[pn] for pn in sorted(list(decay))],
                "weight_decay": weight_decay,
            },
            {
                "params": [param_dict[pn] for pn in sorted(list(no_decay))],
                "weight_decay": 0.0,
            },
        ]
        # new PyTorch nightly has a new 'fused' option for AdamW that is much faster
        use_fused = (device_type == "cuda") and (
            "fused" in inspect.signature(torch.optim.AdamW).parameters
        )
        print(f"using fused AdamW: {use_fused}")
        extra_args = dict(fused=True) if use_fused else dict()
        optimizer = torch.optim.AdamW(
            optim_groups, lr=learning_rate, betas=betas, **extra_args
        )

        return optimizer

    def estimate_mfu(self, fwdbwd_per_iter, dt):
        """estimate model flops utilization (MFU) in units of A100 bfloat16 peak FLOPS"""
        # first estimate the number of flops we do per iteration.
        # see PaLM paper Appendix B as ref: https://arxiv.org/abs/2204.02311
        N = self.get_num_params()
        L, H, Q, T = (
            self.n_layer,
            self.n_head,
            self.n_embd // self.n_head,
            self.block_size,
        )
        flops_per_token = 6 * N + 12 * L * H * Q * T
        flops_per_fwdbwd = flops_per_token * T
        flops_per_iter = flops_per_fwdbwd * fwdbwd_per_iter
        # express our flops throughput as ratio of A100 bfloat16 peak flops
        flops_achieved = flops_per_iter * (1.0 / dt)  # per second
        flops_promised = 312e12  # A100 GPU bfloat16 peak flops is 312 TFLOPS
        mfu = flops_achieved / flops_promised
        return mfu

    @torch.no_grad()
    def generate(self, idx, max_new_tokens, temperature=1.0, top_k=None):
        """
        Take a conditioning sequence of indices idx (LongTensor of shape (b,t)) and complete
        the sequence max_new_tokens times, feeding the predictions back into the model each time.
        Most likely you'll want to make sure to be in model.eval() mode of operation for this.
        """
        for _ in range(max_new_tokens):
            # if the sequence context is growing too long we must crop it at block_size
            idx_cond = (
                idx if idx.size(1) <= self.block_size else idx[:, -self.block_size :]
            )
            # forward the model to get the logits for the index in the sequence
            logits, _, _, _ = self(idx_cond)
            # pluck the logits at the final step and scale by desired temperature
            logits = logits[:, -1, :] / temperature
            # optionally crop the logits to only the top k options
            if top_k is not None:
                v, _ = torch.topk(logits, min(top_k, logits.size(-1)))
                logits[logits < v[:, [-1]]] = -float("Inf")
            # apply softmax to convert logits to (normalized) probabilities
            probs = F.softmax(logits, dim=-1)
            # sample from the distribution
            idx_next = torch.multinomial(probs, num_samples=1)
            # append sampled index to the running sequence and continue
            idx = torch.cat((idx, idx_next), dim=1)

        return idx

    @property
    def init_dict(self):
        """Returns model information in dictionary."""
        init_dict = {
            "n_layer": self.n_layer,
            "vocab_size": self.vocab_size,
            "n_embd": self.n_embd,
            "n_head": self.n_head,
            "dim_feedfwd": self.dim_feedfwd,
            "block_size": self.block_size,
            "dropout": self.dropout,
            "activation": self.activation,
            "layer_norm_eps": self.layer_norm_eps,
            "min_layers": self.min_layers,
            "max_layers": self.max_layers,
            "bias": self.bias,
            "device": self.device,
        }
        return init_dict

    def add_layer(self):
        """Adds a block layer to transformer."""
        if self.n_layer < self.max_layers:
            self.n_layer += 1
            self.recreate_nets()
        # else:
        #     self.add_node()

    def remove_layer(self):
        """Removes a block layer from transformer."""
        if self.n_layer > self.min_layers:
            self.n_layer -= 1
            self.recreate_shrunk_nets()
        # else:
        #     self.add_node()

    def add_node(self, numb_new_nodes=None):
        """Adds nodes to hidden layers of transformer.

        :param numb_new_nodes: Number of nodes to add to hidden layers, defaults to None
        :type numb_new_nodes: int, optional
        """
        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([32, 64, 128], 1)[0]
        self.dim_feedfwd += numb_new_nodes
        self.recreate_nets()
        return {"numb_new_nodes": numb_new_nodes}

    def remove_node(self, numb_new_nodes=None):
        """Removes nodes from hidden layers of transformer.

        :param numb_new_nodes: Number of nodes to remove from hidden layers, defaults to None
        :type numb_new_nodes: int, optional
        """
        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([32, 64, 128], 1)[0]
        self.dim_feedfwd -= numb_new_nodes
        self.recreate_shrunk_nets()
        return {"numb_new_nodes": numb_new_nodes}

    def recreate_nets(self):
        """Recreates neural network."""
        new_transformer = self.create_net()
        self.transformer = self.preserve_parameters(
            old_net=self.transformer, new_net=new_transformer
        )

    def recreate_shrunk_nets(self):
        """Recreates shrunk neural network."""
        new_transformer = self.create_net()
        self.transformer = self.shrink_preserve_parameters(
            old_net=self.transformer, new_net=new_transformer
        )

    def clone(self):
        """Returns clone of neural net with identical parameters."""
        clone = EvolvableGPT(**copy.deepcopy(self.init_dict))
        clone.load_state_dict(self.state_dict())
        return clone

    def preserve_parameters(self, old_net, new_net):
        """Returns new neural network with copied parameters from old network.

        :param old_net: Old neural network
        :type old_net: nn.Module()
        :param new_net: New neural network
        :type new_net: nn.Module()
        """
        old_net_dict = dict(old_net.named_parameters())

        for key, param in new_net.named_parameters():
            if key in old_net_dict.keys():
                if old_net_dict[key].data.size() == param.data.size():
                    param.data = old_net_dict[key].data
                else:
                    if "norm" not in key:
                        old_size = old_net_dict[key].data.size()
                        new_size = param.data.size()
                        if len(param.data.size()) == 1:
                            param.data[: min(old_size[0], new_size[0])] = old_net_dict[
                                key
                            ].data[: min(old_size[0], new_size[0])]
                        else:
                            param.data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                            ] = old_net_dict[key].data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                            ]

        return new_net

    def shrink_preserve_parameters(self, old_net, new_net):
        """Returns shrunk new neural network with copied parameters from old network.

        :param old_net: Old neural network
        :type old_net: nn.Module()
        :param new_net: New neural network
        :type new_net: nn.Module()
        """
        old_net_dict = dict(old_net.named_parameters())

        for key, param in new_net.named_parameters():
            if key in old_net_dict.keys():
                if old_net_dict[key].data.size() == param.data.size():
                    param.data = old_net_dict[key].data
                else:
                    if "norm" not in key:
                        old_size = old_net_dict[key].data.size()
                        new_size = param.data.size()
                        min_0 = min(old_size[0], new_size[0])
                        if len(param.data.size()) == 1:
                            param.data[:min_0] = old_net_dict[key].data[:min_0]
                        else:
                            min_1 = min(old_size[1], new_size[1])
                            param.data[:min_0, :min_1] = old_net_dict[key].data[
                                :min_0, :min_1
                            ]
        return new_net


class LayerNorm(nn.Module):
    """LayerNorm but with an optional bias. PyTorch doesn't support simply bias=False"""

    def __init__(self, ndim, bias, layer_norm_eps=1e-5):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(ndim))
        self.bias = nn.Parameter(torch.zeros(ndim)) if bias else None
        self.layer_norm_eps = layer_norm_eps

    def forward(self, input):
        return F.layer_norm(
            input, self.weight.shape, self.weight, self.bias, self.layer_norm_eps
        )


class CausalSelfAttention(nn.Module):
    def __init__(self, n_embd, n_head, bias, dropout, block_size):
        super().__init__()
        assert n_embd % n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(n_embd, 3 * n_embd, bias=bias)
        # output projection
        self.c_proj = nn.Linear(n_embd, n_embd, bias=bias)
        # regularization
        self.attn_dropout = nn.Dropout(dropout)
        self.resid_dropout = nn.Dropout(dropout)
        self.n_head = n_head
        self.n_embd = n_embd
        self.dropout = dropout
        self.bias = bias
        # flash attention make GPU go brrrrr but support is only in PyTorch >= 2.0
        self.flash = hasattr(torch.nn.functional, "scaled_dot_product_attention")
        if not self.flash:
            print(
                "WARNING: using slow attention. Flash Attention requires PyTorch >= 2.0"
            )
            # causal mask to ensure that attention is only applied to the left in the
            # input sequence
            self.register_buffer(
                "attention_bias",
                torch.tril(torch.ones(block_size, block_size)).view(
                    1, 1, block_size, block_size
                ),
            )

    def forward(self, x, attn_mask=None, layer_past=None, is_causal=True):
        (
            B,
            T,
            C,
        ) = x.size()  # batch size, sequence length, embedding dimensionality (n_embd)

        # calculate query, key, values for all heads in batch and move head
        # forward to be the batch dim
        q, k, v = self.c_attn(x).split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(
            1, 2
        )  # (B, nh, T, hs)

        if layer_past is not None:
            past_key, past_value = layer_past
            k = torch.cat((past_key, k), dim=-2)
            v = torch.cat((past_value, v), dim=-2)

        present = (k, v)

        # causal self-attention; Self-attend: (B, nh, T, hs) x (B, nh, hs, T) ->
        # (B, nh, T, T)
        if self.flash:
            # efficient attention using Flash Attention CUDA kernels
            y = torch.nn.functional.scaled_dot_product_attention(
                q,
                k,
                v,
                attn_mask=attn_mask,
                dropout_p=self.dropout if self.training else 0,
                is_causal=is_causal,
            )
        else:
            # manual implementation of attention
            att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
            att = att.masked_fill(self.attention_bias[:, :, :T, :T] == 0, float("-inf"))
            att = F.softmax(att, dim=-1)
            att = self.attn_dropout(att)
            y = att @ v  # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        # re-assemble all head outputs side by side
        y = y.transpose(1, 2).contiguous().view(B, T, C)

        # output projection
        y = self.resid_dropout(self.c_proj(y))

        return y, present


class Block(nn.Module):
    def __init__(
        self,
        n_embd,
        n_head,
        bias,
        dropout,
        block_size,
        hidden_size,
        activation="GELU",
        layer_norm_eps=1e-5,
    ):
        super().__init__()
        self.ln_1 = LayerNorm(n_embd, bias=bias, layer_norm_eps=layer_norm_eps)
        self.attn = CausalSelfAttention(n_embd, n_head, bias, dropout, block_size)
        self.ln_2 = LayerNorm(n_embd, bias=bias, layer_norm_eps=layer_norm_eps)
        self.mlp = MLP(n_embd, dropout, hidden_size, activation)

    def forward(self, x, attn_mask=None, layer_past=None, is_causal=True):
        attn_output, present = self.attn(
            self.ln_1(x),
            attn_mask=attn_mask,
            layer_past=layer_past,
            is_causal=is_causal,
        )
        x = x + attn_output
        x = x + self.mlp(self.ln_2(x))
        return x, present


class MLP(EvolvableMLP):
    def __init__(self, n_embd, dropout, hidden_size, activation="GELU", **kwargs):
        super().__init__(
            num_inputs=n_embd,
            num_outputs=n_embd,
            hidden_size=[hidden_size],
            layer_norm=False,
            mlp_output_activation=activation,
            **kwargs,
        )
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        """Returns output of neural network.

        :param x: Neural network input
        :type x: torch.Tensor() or np.array
        """
        if not isinstance(x, torch.Tensor):
            x = torch.FloatTensor(np.array(x))
            if self.accelerator is None:
                x = x.to(self.device)
        for value in self.feature_net:
            x = value(x)
        x = self.dropout(x)
        return x

    def get_activation(self, activation_names):
        """Returns activation function for corresponding activation name.

        :param activation_names: Activation function name
        :type activation_names: str
        """
        activation_functions = {
            "Tanh": nn.Tanh,
            "Identity": nn.Identity,
            "ReLU": nn.ReLU,
            "ELU": nn.ELU,
            "Softsign": nn.Softsign,
            "Sigmoid": nn.Sigmoid,
            "Softplus": nn.Softplus,
            "Softmax": nn.Softmax,
            "LeakyReLU": nn.LeakyReLU,
            "PReLU": nn.PReLU,
            "GELU": new_gelu,
        }

        return activation_functions[activation_names]()


class new_gelu(nn.Module):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT).
    Reference: Gaussian Error Linear Units (GELU) paper: https://arxiv.org/abs/1606.08415
    """

    def forward(self, x):
        return (
            0.5
            * x
            * (
                1.0
                + torch.tanh(
                    math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))
                )
            )
        )


class PositionalEncoding(nn.Module):
    """The positional embedding class.
    Converts tensor of input indices into corresponding tensor of position embeddings.
    """

    def __init__(self, max_positions: int, emb_size):
        super().__init__()
        self.embedding = nn.Embedding(max_positions, emb_size)
        self.emb_size = emb_size

    def forward(self, tokens: torch.Tensor):
        """Forward pass through position embedding module.
        :param tokens: Tokens to embed
        :type tokens: torch.Tensor
        """
        return self.embedding(tokens)


class TokenEmbedding(nn.Module):
    """The token embedding class. Converts tensor of input indices into corresponding tensor of token embeddings."""

    def __init__(self, vocab_size: int, emb_size):
        super().__init__()
        self.embedding = nn.Embedding(vocab_size, emb_size)
        self.emb_size = emb_size

    def forward(self, tokens: torch.Tensor):
        """Forward pass through token embedding module.
        :param tokens: Tokens to embed
        :type tokens: torch.Tensor
        """
        # return self.embedding(tokens.long()) * math.sqrt(self.emb_size)
        return self.embedding(tokens)



from typing import Any

import numpy as np


class PettingZooVecEnv:
    """An abstract asynchronous, vectorized environment

    References:
        https://github.com/openai/baselines/tree/master/baselines/common/vec_env
        https://github.com/Farama-Foundation/Gymnasium/blob/main/gymnasium/vector/vector_env.py

    :param num_envs: Number of environments to vectorize
    :type num_envs: int
    :param possible_agents: List of possible agents
    :type possible_agents: list[str]
    """

    metadata: dict[str, Any] = {}
    render_mode: str | None = None
    closed: bool = False

    num_envs: int

    def __init__(self, num_envs, possible_agents):
        self.num_envs = num_envs
        self.agents = possible_agents
        self.num_agents = len(self.agents)

    def reset(self, seed=None, options=None):
        """
        Reset all the environments and return two dictionaries of batched observations and infos.

        :param seed: Random seed, defaults to None
        :type seed: None | int, optional
        :param options: Options dictionary
        :type options: dict[str, Any]
        """
        pass

    def step_async(self, actions):
        """
        Tell all the environments to start taking a step
        with the given actions.
        Call step_wait() to get the results of the step.
        You should not call this if a step_async run is
        already pending.

        :param actions: List of lists of length num_envs, each sub list contains actions for each agent in a given environment
        :type actions: list[list[int | float | np.ndarray]]
        """
        pass

    def step_wait(self):
        """
        Wait for the step taken with step_async().
        """
        pass

    def step(self, actions):
        """Take an action for each parallel environment

        :param actions: Dictionary of vectorized actions for each agent.
        :type actions: dict[str, np.ndarray]
        """
        passed_actions_list = [[] for _ in list(actions.values())[0]]
        for env_idx, _ in enumerate(list(actions.values())[0]):
            for possible_agent in self.agents:
                action = (
                    int(actions[possible_agent][env_idx])
                    if np.isscalar(actions[possible_agent][env_idx])
                    else actions[possible_agent][env_idx]
                )
                passed_actions_list[env_idx].append(action)
        assert (
            len(passed_actions_list) == self.num_envs
        ), "Number of actions passed to the step function must be equal to the number of vectorized environments"
        self.step_async(passed_actions_list)
        return self.step_wait()

    def render(self):
        """Returns the rendered frames from the parallel environments."""
        raise NotImplementedError(
            f"{self.__str__()} render function is not implemented."
        )

    def close(self, **kwargs):
        """
        Clean up the environments' resources.
        """
        if self.closed:
            return

        self.close_extras(**kwargs)
        self.closed = True

    def close_extras(self, **kwargs: Any):
        """Clean up the extra resources e.g. beyond what's in this base class."""
        pass

    @property
    def unwrapped(self):
        """Return the base environment."""
        return self



import os
import time
import warnings
from copy import deepcopy
from datetime import datetime

import numpy as np
import wandb
from torch.utils.data import DataLoader
from tqdm import trange

from agilerl.components.replay_data import ReplayDataset
from agilerl.components.sampler import Sampler


def train_multi_agent(
    env,
    env_name,
    algo,
    pop,
    memory,
    sum_scores=True,
    INIT_HP=None,
    MUT_P=None,
    net_config=None,
    swap_channels=False,
    max_steps=50000,
    evo_steps=25,
    eval_steps=None,
    eval_loop=1,
    learning_delay=0,
    target=None,
    tournament=None,
    mutation=None,
    checkpoint=None,
    checkpoint_path=None,
    overwrite_checkpoints=False,
    save_elite=False,
    elite_path=None,
    wb=False,
    verbose=True,
    accelerator=None,
    wandb_api_key=None,
):
    """The general online multi-agent RL training function. Returns trained population of agents
    and their fitnesses.

    :param env: The environment to train in. Can be vectorized.
    :type env: Gym-style environment
    :param env_name: Environment name
    :type env_name: str
    :param algo: RL algorithm name
    :type algo: str
    :param pop: Population of agents
    :type pop: list[object]
    :param memory: Experience Replay Buffer
    :type memory: object
    :param sum_scores: Boolean flag indicating whether to sum sub-agents scores, typically True for co-operative environments, defaults to True
    :type sum_scores: bool, optional
    :param INIT_HP: Dictionary containing initial hyperparameters.
    :type INIT_HP: dict
    :param MUT_P: Dictionary containing mutation parameters, defaults to None
    :type MUT_P: dict, optional
    :param net_config: Network configuration dictionary, defaults to None
    :type net_config: dict
    :param swap_channels: Swap image channels dimension from last to first
        [H, W, C] -> [C, H, W], defaults to False
    :type swap_channels: bool, optional
    :param max_steps: Maximum number of steps in environment, defaults to 50000
    :type max_steps: int, optional
    :param evo_steps: Evolution frequency (steps), defaults to 25
    :type evo_steps: int, optional
    :param eval_steps: Number of evaluation steps per episode. If None, will evaluate until
        environment terminates or truncates. Defaults to None
    :type eval_steps: int, optional
    :param eval_loop: Number of evaluation episodes, defaults to 1
    :type eval_loop: int, optional
    :param learning_delay: Steps in environment before starting learning, defaults to 0
    :type learning_delay: int, optional
    :param target: Target score for early stopping, defaults to None
    :type target: float, optional
    :param tournament: Tournament selection object, defaults to None
    :type tournament: object, optional
    :param mutation: Mutation object, defaults to None
    :type mutation: object, optional
    :param checkpoint: Checkpoint frequency (steps), defaults to None
    :type checkpoint: int, optional
    :param checkpoint_path: Location to save checkpoint, defaults to None
    :type checkpoint_path: str, optional
    :param overwrite_checkpoints: Overwrite previous checkpoints during training, defaults to False
    :type overwrite_checkpoints: bool, optional
    :param save_elite: Boolean flag indicating whether to save elite member at the end
        of training, defaults to False
    :type save_elite: bool, optional
    :param elite_path: Location to save elite agent, defaults to None
    :type elite_path: str, optional
    :param wb: Weights & Biases tracking, defaults to False
    :type wb: bool, optional
    :param verbose: Display training stats, defaults to True
    :type verbose: bool, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wandb_api_key: API key for Weights & Biases, defaults to None
    :type wandb_api_key: str, optional
    """
    assert isinstance(
        algo, str
    ), "'algo' must be the name of the algorithm as a string."
    assert isinstance(max_steps, int), "Number of steps must be an integer."
    assert isinstance(evo_steps, int), "Evolution frequency must be an integer."
    if target is not None:
        assert isinstance(
            target, (float, int)
        ), "Target score must be a float or an integer."
    if checkpoint is not None:
        assert isinstance(checkpoint, int), "Checkpoint must be an integer."
    assert isinstance(
        wb, bool
    ), "'wb' must be a boolean flag, indicating whether to record run with W&B"
    assert isinstance(verbose, bool), "Verbose must be a boolean."
    if save_elite is False and elite_path is not None:
        warnings.warn(
            "'save_elite' set to False but 'elite_path' has been defined, elite will not\
                      be saved unless 'save_elite' is set to True."
        )
    if checkpoint is None and checkpoint_path is not None:
        warnings.warn(
            "'checkpoint' set to None but 'checkpoint_path' has been defined, checkpoint will not\
                      be saved unless 'checkpoint' is defined."
        )

    start_time = time.time()

    if wb:
        if not hasattr(wandb, "api"):
            if wandb_api_key is not None:
                wandb.login(key=wandb_api_key)
            else:
                warnings.warn("Must login to wandb with API key.")

        config_dict = {}
        if INIT_HP is not None:
            config_dict.update(INIT_HP)
        if MUT_P is not None:
            config_dict.update(MUT_P)
        if net_config is not None:
            config_dict.update(net_config)

        if accelerator is not None:
            accelerator.wait_for_everyone()
            if accelerator.is_main_process:
                wandb.init(
                    # set the wandb project where this run will be logged
                    project="AgileRLMultiAgent",
                    name="{}-MAEvoHPO-{}-{}".format(
                        env_name, algo, datetime.now().strftime("%m%d%Y%H%M%S")
                    ),
                    # track hyperparameters and run metadata
                    config=config_dict,
                )
            accelerator.wait_for_everyone()
        else:
            wandb.init(
                # set the wandb project where this run will be logged
                project="AgileRLMultiAgent",
                name="{}-MAEvoHPO-{}-{}".format(
                    env_name, algo, datetime.now().strftime("%m%d%Y%H%M%S")
                ),
                # track hyperparameters and run metadata
                config=config_dict,
            )
    if accelerator is not None:
        accel_temp_models_path = f"models/{env_name}"
        if accelerator.is_main_process:
            if not os.path.exists(accel_temp_models_path):
                os.makedirs(accel_temp_models_path)

    if hasattr(env, "num_envs"):
        is_vectorised = True
        num_envs = env.num_envs
    else:
        is_vectorised = False
        num_envs = 1

    save_path = (
        checkpoint_path.split(".pt")[0]
        if checkpoint_path is not None
        else "{}-EvoHPO-{}-{}".format(
            env_name, algo, datetime.now().strftime("%m%d%Y%H%M%S")
        )
    )

    if accelerator is not None:
        # Create dataloader from replay buffer
        replay_dataset = ReplayDataset(memory, pop[0].batch_size)
        replay_dataloader = DataLoader(replay_dataset, batch_size=None)
        replay_dataloader = accelerator.prepare(replay_dataloader)
        sampler = Sampler(
            distributed=True, dataset=replay_dataset, dataloader=replay_dataloader
        )
    else:
        sampler = Sampler(distributed=False, memory=memory)

    if accelerator is not None:
        print(f"\nDistributed training on {accelerator.device}...")
    else:
        print("\nTraining...")

    bar_format = "{l_bar}{bar:10}| {n:4}/{total_fmt} [{elapsed:>7}<{remaining:>7}, {rate_fmt}{postfix}]"
    if accelerator is not None:
        pbar = trange(
            max_steps,
            unit="step",
            bar_format=bar_format,
            ascii=True,
            disable=not accelerator.is_local_main_process,
        )
    else:
        pbar = trange(max_steps, unit="step", bar_format=bar_format, ascii=True)

    agent_ids = deepcopy(env.agents)
    pop_actor_loss = [{agent_id: [] for agent_id in agent_ids} for _ in pop]
    pop_critic_loss = [{agent_id: [] for agent_id in agent_ids} for _ in pop]
    pop_fitnesses = []
    total_steps = 0
    loss = None
    checkpoint_count = 0

    # Pre-training mutation
    if accelerator is None:
        if mutation is not None:
            pop = mutation.mutation(pop, pre_training_mut=True)
    # RL training loop
    while np.less([agent.steps[-1] for agent in pop], max_steps).all():
        if accelerator is not None:
            accelerator.wait_for_everyone()
        pop_episode_scores = []
        for agent_idx, agent in enumerate(pop):  # Loop through population
            state, info = env.reset()  # Reset environment at start of episode
            scores = (
                np.zeros((num_envs, 1))
                if sum_scores
                else np.zeros((num_envs, len(agent_ids)))
            )
            losses = {agent_id: [] for agent_id in agent_ids}
            completed_episode_scores = []
            steps = 0

            if swap_channels:
                if not is_vectorised:
                    state = {
                        agent_id: np.moveaxis(np.expand_dims(s, 0), [-1], [-3])
                        for agent_id, s in state.items()
                    }
                else:
                    state = {
                        agent_id: np.moveaxis(s, [-1], [-3])
                        for agent_id, s in state.items()
                    }
            for idx_step in range(evo_steps // num_envs):
                # Get next action from agent
                cont_actions, discrete_action = agent.get_action(
                    states=state, training=True, infos=info
                )
                if agent.discrete_actions:
                    action = discrete_action
                else:
                    action = cont_actions

                if not is_vectorised:
                    action = {agent: act[0] for agent, act in action.items()}
                    cont_actions = {
                        agent: act[0] for agent, act in cont_actions.items()
                    }

                # Act in environment
                next_state, reward, termination, truncation, info = env.step(action)
                score_increment = (
                    (
                        np.sum(np.array(list(reward.values())).transpose(), axis=-1)[
                            :, np.newaxis
                        ]
                        if is_vectorised
                        else np.sum(
                            np.array(list(reward.values())).transpose(), axis=-1
                        )
                    )
                    if sum_scores
                    else np.array(list(reward.values())).transpose()
                )

                scores += score_increment
                total_steps += num_envs
                steps += num_envs
                # Save experience to replay buffer
                if swap_channels:
                    if not is_vectorised:
                        state = {
                            agent_id: np.squeeze(s) for agent_id, s in state.items()
                        }
                    next_state = {
                        agent_id: np.moveaxis(ns, [-1], [-3])
                        for agent_id, ns in next_state.items()
                    }

                memory.save_to_memory(
                    state,
                    cont_actions,
                    reward,
                    next_state,
                    termination,
                    is_vectorised=is_vectorised,
                )

                # Learn according to learning frequency
                # Handle learn steps > num_envs
                if agent.learn_step > num_envs:
                    learn_step = agent.learn_step // num_envs
                    if (
                        idx_step % learn_step == 0
                        and len(memory) >= agent.batch_size
                        and memory.counter > learning_delay
                    ):
                        # Sample replay buffer
                        experiences = sampler.sample(agent.batch_size)
                        # Learn according to agent's RL algorithm
                        loss = agent.learn(experiences)
                        for agent_id in agent_ids:
                            losses[agent_id].append(loss[agent_id])

                # Handle num_envs > learn step; learn multiple times per step in env
                elif (
                    len(memory) >= agent.batch_size and memory.counter > learning_delay
                ):
                    for _ in range(num_envs // agent.learn_step):
                        # Sample replay buffer
                        experiences = sampler.sample(agent.batch_size)
                        # Learn according to agent's RL algorithm
                        loss = agent.learn(experiences)
                        for agent_id in agent_ids:
                            losses[agent_id].append(loss[agent_id])

                # Update the state
                if swap_channels and not is_vectorised:
                    next_state = {
                        agent_id: np.expand_dims(ns, 0)
                        for agent_id, ns in next_state.items()
                    }

                state = next_state

                reset_noise_indices = []
                # Find which agents are "done" - i.e. terminated or truncated
                dones = {
                    agent_id: termination[agent_id] | truncation[agent_id]
                    for agent_id in agent.agent_ids
                }
                if not is_vectorised:
                    dones = {agent: np.array([done]) for agent, done in dones.items()}

                for idx, agent_dones in enumerate(zip(*dones.values())):
                    if all(agent_dones):
                        completed_score = (
                            float(scores[idx]) if sum_scores else list(scores[idx])
                        )
                        completed_episode_scores.append(completed_score)
                        agent.scores.append(completed_score)
                        scores[idx].fill(0)
                        reset_noise_indices.append(idx)
                        if not is_vectorised:
                            state, info = env.reset()
                agent.reset_action_noise(reset_noise_indices)
            pbar.update(evo_steps // len(pop))

            agent.steps[-1] += steps
            pop_episode_scores.append(completed_episode_scores)
            if len(losses[agent_ids[0]]) > 0:
                if all([losses[a_id] for a_id in agent_ids]):
                    for agent_id in agent_ids:
                        actor_losses, critic_losses = list(zip(*losses[agent_id]))
                        actor_losses = [
                            loss for loss in actor_losses if loss is not None
                        ]
                        if actor_losses:
                            pop_actor_loss[agent_idx][agent_id].append(
                                np.mean(actor_losses)
                            )
                        pop_critic_loss[agent_idx][agent_id].append(
                            np.mean(critic_losses)
                        )
        # Evaluate population
        fitnesses = [
            agent.test(
                env,
                swap_channels=swap_channels,
                max_steps=eval_steps,
                loop=eval_loop,
                sum_scores=sum_scores,
            )
            for agent in pop
        ]
        pop_fitnesses.append(fitnesses)
        if sum_scores:
            mean_scores = [
                (
                    np.mean(episode_scores)
                    if len(episode_scores) > 0
                    else "0 completed episodes"
                )
                for episode_scores in pop_episode_scores
            ]
            mean_score_dict = {
                "train/mean_score": np.mean(
                    [
                        mean_score
                        for mean_score in mean_scores
                        if not isinstance(mean_score, str)
                    ]
                )
            }
            fitness_dict = {
                "eval/mean_fitness": np.mean(fitnesses),
                "eval/best_fitness": np.max(fitnesses),
            }
        else:
            pop_mean_scores = [
                np.mean(np.array(score), axis=0)
                for score in pop_episode_scores
                if score
            ]
            if pop_episode_scores:
                mean_scores = np.stack(pop_mean_scores, axis=0)
                mean_score_dict = {
                    "train/mean_score/" + agent: np.mean(mean_scores[:, idx], axis=-1)
                    for idx, agent in enumerate(agent_ids)
                }
            else:
                mean_score_dict = {
                    "train/mean_score/" + agent: np.nan
                    for idx, agent in enumerate(agent_ids)
                }
            mean_fitnesses = np.mean(fitnesses, axis=0)
            max_fitnesses = np.max(fitnesses, axis=0)
            fitness_dict = {
                "eval/mean_fitness/" + agent: mean_fitnesses[idx]
                for idx, agent in enumerate(agent_ids)
            }
            best_fitness_dict = {
                "eval/best_fitness/" + agent: max_fitnesses[idx]
                for idx, agent in enumerate(agent_ids)
            }
            fitness_dict.update(best_fitness_dict)

        if wb:
            wandb_dict = {
                "global_step": (
                    total_steps * accelerator.state.num_processes
                    if accelerator is not None and accelerator.is_main_process
                    else total_steps
                ),
            }
            wandb_dict.update(fitness_dict)
            wandb_dict.update(mean_score_dict)

            actor_loss_dict = {}
            critic_loss_dict = {}

            for agent_idx, agent in enumerate(pop):
                for agent_id, actor_loss, critic_loss in zip(
                    pop_actor_loss[agent_idx].keys(),
                    pop_actor_loss[agent_idx].values(),
                    pop_critic_loss[agent_idx].values(),
                ):
                    if actor_loss:
                        actor_loss_dict[
                            f"train/agent_{agent_idx}_{agent_id}_actor_loss"
                        ] = np.mean(actor_loss[-10:])
                        wandb_dict.update(actor_loss_dict)
                    if critic_loss:
                        critic_loss_dict[
                            f"train/agent_{agent_idx}_{agent_id}_critic_loss"
                        ] = np.mean(critic_loss[-10:])
                        wandb_dict.update(critic_loss_dict)

            if accelerator is not None:
                accelerator.wait_for_everyone()
                if accelerator.is_main_process:
                    wandb.log(wandb_dict)
                accelerator.wait_for_everyone()
            else:
                wandb.log(wandb_dict)

            for idx, agent in enumerate(pop):
                wandb.log(
                    {
                        f"learn_step_agent_{idx}": agent.learn_step,
                        f"learning_rate_actor_agent_{idx}": agent.lr_actor,
                        f"learning_rate_critic_agent_{idx}": agent.lr_critic,
                        f"batch_size_agent_{idx}": agent.batch_size,
                        f"indi_fitness_agent_{idx}": agent.fitness[-1],
                    }
                )
        # Update step counter
        for agent in pop:
            agent.steps.append(agent.steps[-1])

        # Early stop if consistently reaches target
        if target is not None:
            if (
                np.all(
                    np.greater([np.mean(agent.fitness[-10:]) for agent in pop], target)
                )
                and len(pop[0].steps) >= 100
            ):
                if wb:
                    wandb.finish()
                return pop, pop_fitnesses

        # Tournament selection and population mutation
        if tournament and mutation is not None:
            if accelerator is not None:
                accelerator.wait_for_everyone()
                for model in pop:
                    model.unwrap_models()
                accelerator.wait_for_everyone()
                if accelerator.is_main_process:
                    elite, pop = tournament.select(pop)
                    pop = mutation.mutation(pop)
                    for pop_i, model in enumerate(pop):
                        model.save_checkpoint(
                            f"{accel_temp_models_path}/{algo}_{pop_i}.pt"
                        )
                accelerator.wait_for_everyone()
                if not accelerator.is_main_process:
                    for pop_i, model in enumerate(pop):
                        model.load_checkpoint(
                            f"{accel_temp_models_path}/{algo}_{pop_i}.pt"
                        )
                accelerator.wait_for_everyone()
                for model in pop:
                    model.wrap_models()
            else:
                elite, pop = tournament.select(pop)
                pop = mutation.mutation(pop)

            if save_elite:
                elite_save_path = (
                    elite_path.split(".pt")[0]
                    if elite_path is not None
                    else f"{env_name}-elite_{algo}"
                )
                elite.save_checkpoint(f"{elite_save_path}.pt")

        if verbose:
            if sum_scores:
                fitness = ["%.2f" % fitness for fitness in fitnesses]
                avg_fitness = ["%.2f" % np.mean(agent.fitness[-5:]) for agent in pop]
                avg_score = ["%.2f" % np.mean(agent.scores[-10:]) for agent in pop]
            else:
                fitness_arr = np.array([fitness for fitness in fitnesses])
                avg_fitness_arr = np.array(
                    [np.mean(agent.fitness[-5:], axis=0) for agent in pop]
                )
                avg_score_arr = np.array(
                    [np.mean(agent.scores[-10:], axis=0) for agent in pop]
                )
                fitness = {
                    agent: fitness_arr[:, idx] for idx, agent in enumerate(agent_ids)
                }
                avg_fitness = {
                    agent: avg_fitness_arr[idx] for idx, agent in enumerate(agent_ids)
                }
                avg_score = {
                    agent: avg_score_arr[idx] for idx, agent in enumerate(agent_ids)
                }
                mean_scores = {
                    agent: mean_scores[:, idx] for idx, agent in enumerate(agent_ids)
                }
            agents = [agent.index for agent in pop]
            num_steps = [agent.steps[-1] for agent in pop]
            muts = [agent.mut for agent in pop]
            pbar.update(0)

            print()
            print(
                "DateTime, now, H:m:s-u",
                datetime.now().hour,
                ":",
                datetime.now().minute,
                ":",
                datetime.now().second,
                "-",
                datetime.now().microsecond,
            )
            total_time = time.time() - start_time
            print(
                "Steps",
                total_steps / total_time,
                "per sec,",
                total_steps / (total_time / 60),
                "per min.",
            )
            print(
                f"""
                --- Global Steps {total_steps} ---
                Fitness:\t{fitness}
                Score:\t\t{mean_scores}
                5 fitness avgs:\t{avg_fitness}
                10 score avgs:\t{avg_score}
                Agents:\t\t{agents}
                Steps:\t\t{num_steps}
                Mutations:\t{muts}
                """,
                end="\r",
            )

        # Save model checkpoint
        if checkpoint is not None:
            if pop[0].steps[-1] // checkpoint > checkpoint_count:
                if accelerator is not None:
                    accelerator.wait_for_everyone()
                    for model in pop:
                        model.unwrap_models()
                    accelerator.wait_for_everyone()
                    if accelerator.is_main_process:
                        for i, agent in enumerate(pop):
                            current_checkpoint_path = (
                                f"{save_path}_{i}.pt"
                                if overwrite_checkpoints
                                else f"{save_path}_{i}_{agent.steps[-1]}.pt"
                            )
                            agent.save_checkpoint(current_checkpoint_path)
                        print("Saved checkpoint.")
                    accelerator.wait_for_everyone()
                    for model in pop:
                        model.wrap_models()
                    accelerator.wait_for_everyone()
                else:
                    for i, agent in enumerate(pop):
                        current_checkpoint_path = (
                            f"{save_path}_{i}.pt"
                            if overwrite_checkpoints
                            else f"{save_path}_{i}_{agent.steps[-1]}.pt"
                        )
                        agent.save_checkpoint(current_checkpoint_path)
                    print("Saved checkpoint.")
                checkpoint_count += 1

    if wb:
        if accelerator is not None:
            accelerator.wait_for_everyone()
            if accelerator.is_main_process:
                wandb.finish()
            accelerator.wait_for_everyone()
        else:
            wandb.finish()

    pbar.close()
    return pop, pop_fitnesses



import os

import h5py
import minari
from minari.storage.datasets_root_dir import get_dataset_path
from minari.storage.hosting import download_dataset
from minari.storage.local import load_dataset


def load_minari_dataset(dataset_id, accelerator=None, remote=False):
    if remote:
        if dataset_id not in list(minari.list_remote_datasets().keys()):
            raise KeyError(
                "Enter a valid remote Minari Dataset ID. check https://minari.farama.org/ for more details."
            )

    file_path = get_dataset_path(dataset_id)

    if not os.path.exists(file_path):
        if remote:
            if accelerator is not None:
                accelerator.wait_for_everyone()
                if accelerator.is_main_process:
                    print("download dataset: ", dataset_id)
                    download_dataset(dataset_id)
                accelerator.wait_for_everyone()
            else:
                print("download dataset: ", dataset_id)
                download_dataset(dataset_id)
        else:
            raise FileNotFoundError(
                f"No local Dataset found for dataset id {dataset_id}. check https://minari.farama.org/ for more details on remote dataset. For loading a remote dataset assign remote=True"
            )

    minari_dataset = load_dataset(dataset_id)

    return minari_dataset


def minari_to_agile_buffer(dataset_id, memory, accelerator=None, remote=False):
    minari_dataset = load_minari_dataset(dataset_id, accelerator, remote)

    for episode in minari_dataset.iterate_episodes():
        for num_steps in range(0, len(episode.rewards)):
            observation = episode.observations[num_steps]
            next_observation = episode.observations[num_steps + 1]
            action = episode.actions[num_steps]
            reward = episode.rewards[num_steps]
            terminal = episode.terminations[num_steps]
            memory.save_to_memory(
                observation, action, reward, next_observation, terminal
            )

    return memory


def minari_to_agile_dataset(dataset_id, remote=False):
    observations = []
    next_observations = []
    actions = []
    rewards = []
    terminals = []

    dataset = load_minari_dataset(dataset_id, remote)

    for episode in dataset.iterate_episodes():
        observations.extend(episode.observations[:-1])
        next_observations.extend(episode.observations[1:])
        actions.extend(episode.actions[:])
        rewards.extend(episode.rewards[:])
        terminals.extend(episode.terminations[:])

    agile_dataset_id = dataset_id.split("-")
    agile_dataset_id[0] = agile_dataset_id[0] + "_agile"
    agile_dataset_id = "-".join(agile_dataset_id)

    agile_file_path = get_dataset_path(agile_dataset_id)

    agile_dataset_path = os.path.join(agile_file_path, "data")
    os.makedirs(agile_dataset_path, exist_ok=True)
    data_path = os.path.join(agile_dataset_path, "main_data.hdf5")

    # with h5py.File(os.path.join(agile_file_path, "data", "main_data.hdf5"), "w") as f:
    f = h5py.File(data_path, "w")

    f.create_dataset("observations", data=observations)
    f.create_dataset("next_observations", data=next_observations)
    f.create_dataset("actions", data=actions)
    f.create_dataset("rewards", data=rewards)
    f.create_dataset("terminals", data=terminals)

    return f



import random
from collections import deque, namedtuple

import numpy as np
import torch

from agilerl.components.segment_tree import MinSegmentTree, SumSegmentTree


class ReplayBuffer:
    """The Experience Replay Buffer class. Used to store experiences and allow
    off-policy learning.

    :param memory_size: Maximum length of replay buffer
    :type memory_size: int
    :param field_names: Field names for experience named tuple, e.g. ['state', 'action', 'reward']
    :type field_names: list[str]
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to None
    :type device: str, optional
    """

    def __init__(self, memory_size, field_names, device=None):
        assert memory_size > 0, "Mmeory size must be greater than zero."
        assert len(field_names) > 0, "Field names must contain at least one field name."

        self.memory_size = memory_size
        self.memory = deque(maxlen=memory_size)
        self.field_names = field_names
        self.experience = namedtuple("Experience", field_names=self.field_names)
        self.counter = 0  # update cycle counter
        self.device = device

    def __len__(self):
        return len(self.memory)

    def _add(self, *args):
        """Adds experience to memory."""
        e = self.experience(*args)
        self.memory.append(e)

    def _process_transition(self, experiences, np_array=False):
        """Returns transition dictionary from experiences."""
        transition = {}
        for field in self.field_names:
            ts = [
                np.expand_dims(getattr(e, field), axis=0)
                for e in experiences
                if e is not None
            ]
            ts = np.vstack(ts)

            # Handle numpy stacking
            if field in [
                "done",
                "termination",
                "terminated",
                "truncation",
                "truncated",
            ]:
                ts = ts.astype(np.uint8)

            if not np_array:
                # Handle torch tensor creation
                ts = torch.from_numpy(ts).float()

                # Place on device
                if self.device is not None:
                    ts = ts.to(self.device)

            transition[field] = ts
        return transition

    def sample(self, batch_size, return_idx=False):
        """Returns sample of experiences from memory.

        :param batch_size: Number of samples to return
        :type batch_size: int
        :param return_idx: Boolean flag to return index of samples randomly selected, defaults to False
        :type return_idx: bool
        """
        if return_idx:
            idxs = np.random.choice(len(self.memory), size=batch_size, replace=False)
            experiences = list(map(lambda i: self.memory[i], idxs))
            transition = self._process_transition(experiences)
            transition["idxs"] = idxs
        else:
            experiences = random.sample(self.memory, k=batch_size)
            transition = self._process_transition(experiences)

        return tuple(transition.values())

    def save_to_memory_single_env(self, *args):
        """Saves experience to memory.

        :param *args: Variable length argument list. Contains transition elements in consistent order,
            e.g. state, action, reward, next_state, done
        """
        self._add(*args)
        self.counter += 1

    def save_to_memory_vect_envs(self, *args):
        """Saves multiple experiences to memory.

        :param *args: Variable length argument list. Contains batched transition elements in consistent order,
            e.g. states, actions, rewards, next_states, dones
        """
        for transition in zip(*args):
            self._add(*transition)
            self.counter += 1

    def save_to_memory(self, *args, is_vectorised=False):
        """Applies appropriate save_to_memory function depending on whether
        the environment is vectorised or not.

        :param *args: Variable length argument list. Contains batched or unbatched transition elements in consistent order,
            e.g. states, actions, rewards, next_states, dones
        :param is_vectorised: Boolean flag indicating if the environment has been vectorised
        :type is_vectorised: bool
        """
        if is_vectorised:
            self.save_to_memory_vect_envs(*args)
        else:
            self.save_to_memory_single_env(*args)


class MultiStepReplayBuffer(ReplayBuffer):
    """The Multi-step Experience Replay Buffer class. Used to store experiences and allow
    off-policy learning.

    :param memory_size: Maximum length of replay buffer
    :type memory_size: int
    :param field_names: Field names for experience named tuple, e.g. ['state', 'action', 'reward']
    :type field_names: list[str]
    :param num_envs: Number of parallel environments for training
    :type num_envs: int
    :param n_step: Step number to calculate n-step td error, defaults to 3
    :type n_step: int, optional
    :param gamma: Discount factor, defaults to 0.99
    :type gamma: float, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to None
    :type device: str, optional
    """

    def __init__(
        self,
        memory_size,
        field_names,
        num_envs,
        n_step=3,
        gamma=0.99,
        device=None,
    ):
        super().__init__(memory_size, field_names, device)
        assert (
            "reward" in field_names
        ), "Reward must be saved in replay buffer under the field name 'reward'."
        assert (
            "next_state" in field_names
        ), "Next state must be saved in replay buffer under the field name 'next_state'."
        assert (
            "done" in field_names
            or "termination" in field_names
            or "terminated" in field_names
        ), "Done/termination must be saved in replay buffer under the field name 'done', 'termination', or 'terminated'."
        self.num_envs = num_envs
        self.n_step_buffers = [deque(maxlen=n_step) for i in range(num_envs)]
        self.args_deque = deque(maxlen=n_step)
        self.n_step = n_step
        self.gamma = gamma

    def save_to_memory_single_env(self, *args):
        """Saves experience to memory.

        :param *args: Variable length argument list. Contains transition elements in consistent order,
            e.g. state, action, reward, next_state, done
        """
        self.args_deque.append(args)
        transition = self.experience(*args)
        self.n_step_buffers[0].append(transition)

        # single step transition is not ready
        if len(self.n_step_buffers[0]) < self.n_step:
            return ()

        # make a n-step transition
        args = self._get_n_step_info(self.n_step_buffers[0], self.gamma)
        self._add(*args)
        self.counter += 1

        return self.args_deque[0]

    def save_to_memory_vect_envs(self, *args):
        """Saves multiple experiences to memory.

        :param *args: Variable length argument list. Contains transition elements in consistent order,
            e.g. state, action, reward, next_state, done
        """
        self.args_deque.append(args)
        for buffer, *transition in zip(self.n_step_buffers, *args):
            transition = self.experience(*transition)
            buffer.append(transition)

        # single step transition is not ready
        if any(len(buffer) < self.n_step for buffer in self.n_step_buffers):
            return ()
        else:
            for buffer in self.n_step_buffers:
                # make a n-step transition
                single_step_args = self._get_n_step_info(buffer, self.gamma)
                self._add(*single_step_args)
                self.counter += 1

            return self.args_deque[0]

    def sample_from_indices(self, idxs):
        """Returns sample of experiences from memory using provided indices.

        :param idxs: Indices to sample
        :type idxs: list[int]
        """
        experiences = list(map(lambda i: self.memory[i], idxs))
        transition = self._process_transition(experiences)
        return tuple(transition.values())

    def _get_n_step_info(self, n_step_buffer, gamma):
        """Returns n step reward, next_state, and done, as well as other saved transition elements, in order."""
        # info of the last transition
        # t = [n_step_buffer[0]]
        t = [n_step_buffer[0]]
        transition = self._process_transition(t, np_array=True)

        vect_reward = transition["reward"][0]
        vect_next_state = transition["next_state"][0]
        if "done" in transition.keys():
            vect_done = transition["done"][0]
        elif "termination" in transition.keys():
            vect_done = transition["termination"][0]
        else:
            vect_done = transition["terminated"][0]

        for idx, ts in enumerate(list(n_step_buffer)[1:]):
            if not vect_done:
                vect_r, vect_n_s = (ts.reward, ts.next_state)

                if "done" in transition.keys():
                    vect_d = ts.done
                elif "termination" in transition.keys():
                    vect_d = ts.termination
                else:
                    vect_d = ts.terminated

                vect_reward += vect_r * gamma ** (idx + 1)
                vect_done = np.array([vect_d])
                vect_next_state = vect_n_s

        transition["reward"] = vect_reward
        transition["next_state"] = vect_next_state
        if "done" in transition.keys():
            transition["done"] = vect_done
        elif "termination" in transition.keys():
            transition["termination"] = vect_done
        else:
            transition["terminated"] = vect_done
        transition["state"] = transition["state"][0]
        transition["action"] = transition["action"][0]

        return tuple(transition.values())


class PrioritizedReplayBuffer(MultiStepReplayBuffer):
    """The Prioritized Experience Replay Buffer class. Used to store experiences and allow
    off-policy learning.

    :param memory_size: Maximum length of replay buffer
    :type memory_size: int
    :param field_names: Field names for experience named tuple, e.g. ['state', 'action', 'reward']
    :type field_names: list[str]
    :param num_envs: Number of parallel environments for training
    :type num_envs: int
    :param alpha: Alpha parameter for prioritized replay buffer, defaults to 0.6
    :type alpha: float, optional
    :param n_step: Step number to calculate n-step td error, defaults to 1
    :type n_step: int, optional
    :param gamma: Discount factor, defaults to 0.99
    :type gamma: float, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to None
    :type device: str, optional
    """

    def __init__(
        self,
        memory_size,
        field_names,
        num_envs,
        alpha=0.6,
        n_step=1,
        gamma=0.99,
        device=None,
    ):
        super().__init__(memory_size, field_names, num_envs, n_step, gamma, device)
        self.max_priority, self.tree_ptr = 1.0, 0
        self.alpha = alpha

        # capacity must be positive and a power of 2.
        tree_capacity = 1
        while tree_capacity < self.memory_size:
            tree_capacity *= 2

        self.sum_tree = SumSegmentTree(tree_capacity)
        self.min_tree = MinSegmentTree(tree_capacity)

    def _add(self, *args):
        super()._add(*args)
        self.sum_tree[self.tree_ptr] = self.max_priority**self.alpha
        self.min_tree[self.tree_ptr] = self.max_priority**self.alpha
        self.tree_ptr = (self.tree_ptr + 1) % self.memory_size

    def sample(self, batch_size, beta=0.4):
        """Returns sample of experiences from memory.

        :param batch_size: Number of samples to return
        :type batch_size: int
        """
        idxs = self._sample_proprtional(batch_size)
        experiences = [self.memory[i] for i in idxs]
        transition = self._process_transition(experiences)

        weights = torch.from_numpy(
            np.array([self._calculate_weight(i, beta) for i in idxs])
        ).float()

        if self.device is not None:
            weights = weights.to(self.device)

        transition["weights"] = weights
        transition["idxs"] = idxs

        return tuple(transition.values())

    def update_priorities(self, idxs, priorities):
        """Update priorities of sampled transitions."""
        for idx, priority in zip(idxs, priorities):
            self.sum_tree[idx] = priority**self.alpha
            self.min_tree[idx] = priority**self.alpha
            self.max_priority = max(self.max_priority, priority)

    def _sample_proprtional(self, batch_size):
        """Sample indices based on proportions.

        :param batch_size: Sample size
        :type batch_size: int
        """
        idxs = []
        p_total = self.sum_tree.sum(0, len(self) - 1)
        segment = p_total / batch_size
        for i in range(batch_size):
            a = segment * i
            b = segment * (i + 1)
            upperbound = random.uniform(a, b)
            idx = self.sum_tree.retrieve(upperbound)
            idxs.append(idx)
        return idxs

    def _calculate_weight(self, idx, beta):
        """Calculate the weight of the experience at idx."""
        # get max weight
        p_min = self.min_tree.min() / self.sum_tree.sum()
        max_weight = (p_min * len(self)) ** (-beta)

        # calculate weights
        p_sample = self.sum_tree[idx] / self.sum_tree.sum()
        weight = (p_sample * len(self)) ** (-beta)
        weight = weight / max_weight

        return weight



from __future__ import annotations

import os
import pickle as pkl
from typing import Any


class Cache:
    def __init__(self, cache_init: dict | None = None) -> None:
        assert cache_init is None or isinstance(cache_init, dict)
        if cache_init is None:
            cache_init = {}
        self.cache = cache_init
        self.cache_hit_rate = 1.0

    def dump(self, file_name: str):
        if not os.path.exists(os.path.dirname(file_name)):
            os.makedirs(os.path.dirname(file_name))
        with open(file_name, "wb") as f:
            pkl.dump(self.cache, f)

    def load(self, file_name: str):
        with open(file_name, "rb") as f:
            self.cache.update(pkl.load(f))

    def __getitem__(self, key: str) -> dict:
        self.cache_hit_rate = (self.cache_hit_rate * 0.99) + 0.01
        return self.cache[key]

    def __setitem__(self, key: str, newvalue: Any):
        self.cache_hit_rate = self.cache_hit_rate * 0.99
        self.cache[key] = newvalue

    def __contains__(self, key: str) -> bool:
        return key in self.cache

    def __len__(self) -> int:
        return len(self.cache)

    def items(self):
        return list(self.cache.items())

    def keys(self):
        return list(self.cache.keys())

    def values(self):
        return list(self.cache.values())

    def update(self, new_stuff: dict):
        self.cache.update(new_stuff)

    def get_hit_rate(self):
        return self.cache_hit_rate

    def get_cache(self):
        return self.cache



import math
from typing import Any, Callable, List, Union

import numpy as np
import torch
import torch.nn as nn


def map_pytree(f: Callable[[Union[np.ndarray, torch.Tensor]], Any], item: Any):
    if isinstance(item, dict):
        return {k: map_pytree(f, v) for k, v in item.items()}
    elif isinstance(item, list) or isinstance(item, set) or isinstance(item, tuple):
        return [map_pytree(f, v) for v in item]
    elif isinstance(item, np.ndarray) or isinstance(item, torch.Tensor):
        return f(item)
    else:
        return item


def to(item: Any, device: torch.device):
    return map_pytree(lambda x: torch.tensor(x).to(device), item)


def to_decorator(f, device):
    def new_f(*args, **kwargs):
        return to(f(*args, **kwargs), device)

    return new_f


def parameter_norm(model: nn.Module):
    norm = 0.0
    for param in model.parameters():
        norm += (param.norm() ** 2).item()
    return math.sqrt(norm)


def get_transformer_logs(
    attentions: List[torch.Tensor], model: nn.Module, attn_mask: torch.Tensor
):
    logs = {}
    n = attn_mask.sum()
    model_attention_entropy = -sum(
        map(
            lambda x: ((x * torch.log(x + 1e-7)).sum(dim=-1) * attn_mask.unsqueeze(1))
            .sum()
            .item(),
            attentions,
        )
    ) / (len(attentions) * n)
    model_parameter_norm = parameter_norm(model)
    logs["attention_entropy"] = (model_attention_entropy, n * len(attentions))
    logs["parameter_norm"] = (model_parameter_norm, 1)
    return logs



import numpy as np


class TournamentSelection:
    """The tournament selection class.

    :param tournament_size: Tournament selection size
    :type tournament_size: int
    :param elitism: Elitism in tournament selection
    :type elitism: bool
    :param population_size: Number of agents in population
    :type population_size: int
    :param eval_loop: Number of most recent fitness scores to use in evaluation
    :type eval_loop: int
    """

    def __init__(self, tournament_size, elitism, population_size, eval_loop):
        assert tournament_size > 0, "Tournament size must be greater than zero."
        assert isinstance(elitism, bool), "Elitism must be boolean value True or False."
        assert population_size > 0, "Population size must be greater than zero."
        assert eval_loop > 0, "Evo step must be greater than zero."

        self.tournament_size = tournament_size
        self.elitism = elitism
        self.population_size = population_size
        self.eval_loop = eval_loop

    def _tournament(self, fitness_values):
        selection = np.random.randint(0, len(fitness_values), size=self.tournament_size)
        selection_values = [fitness_values[i] for i in selection]
        winner = selection[np.argmax(selection_values)]
        return winner

    def _elitism(self, population):
        """Returns elite member of population and its id."""
        last_fitness = [np.mean(indi.fitness[-self.eval_loop :]) for indi in population]
        rank = np.argsort(last_fitness).argsort()

        max_id = max([ind.index for ind in population])

        model = population[np.argsort(rank)[-1]]
        elite = model.clone()
        return elite, rank, max_id

    def select(self, population):
        """Returns best agent and new population of agents following tournament selection.

        :param population: Population of agents
        :type population: list[object]
        """
        elite, rank, max_id = self._elitism(population)

        new_population = []
        if self.elitism:  # keep top agent in population
            new_population.append(elite.clone(wrap=False))
            selection_size = self.population_size - 1
        else:
            selection_size = self.population_size

        # select parents of next gen using tournament selection
        for idx in range(selection_size):
            max_id += 1
            actor_parent = population[self._tournament(rank)]
            new_individual = actor_parent.clone(max_id, wrap=False)
            new_population.append(new_individual)

        return elite, new_population



from collections import OrderedDict

import torch
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer


def unwrap_optimizer(optimizer, network, lr):
    if isinstance(optimizer, AcceleratedOptimizer):
        if isinstance(network, (list, tuple)):
            optim_arg = [{"params": net.parameters(), "lr": lr} for net in network]
            unwrapped_optimizer = type(optimizer.optimizer)(optim_arg)
        else:
            unwrapped_optimizer = type(optimizer.optimizer)(network.parameters(), lr=lr)
        unwrapped_optimizer.load_state_dict(optimizer.state_dict())
        return unwrapped_optimizer
    else:
        return optimizer


def chkpt_attribute_to_device(chkpt_dict, device):
    """Place checkpoint attributes on device. Used when loading saved agents.

    :param chkpt_dict: Checkpoint dictionary
    :type chkpt_dict: dict
    :param device: Device for accelerated computing, 'cpu' or 'cuda'
    :type device: str
    """
    for key, value in chkpt_dict.items():
        if hasattr(value, "device") and not isinstance(value, Accelerator):
            chkpt_dict[key] = value.to(device)
    return chkpt_dict


def key_in_nested_dict(nested_dict, target):
    """Helper function to determine if key is in nested dictionary

    :param nested_dict: Nested dictionary
    :type nested_dict: Dict[str, Dict[str, ...]]
    :param target: Target string
    :type target: str
    """
    for k, v in nested_dict.items():
        if k == target:
            return True
        if isinstance(v, dict):
            return key_in_nested_dict(v, target)
    return False


def compile_model(model, mode: str | None = "default"):
    """Compiles torch model if not already compiled

    :param model: torch model
    :type model: nn.Module
    :param mode: torch compile mode, defaults to "default"
    :type mode: str, optional
    :return: compiled model
    :rtype: OptimizedModule
    """
    return (
        torch.compile(model, mode=mode)
        if not isinstance(model, torch._dynamo.eval_frame.OptimizedModule)
        and mode is not None
        else model
    )


def remove_compile_prefix(state_dict):
    """Removes _orig_mod prefix on state dict created by torch compile

    :param state_dict: model state dict
    :type state_dict: dict
    :return: state dict with prefix removed
    :rtype: dict
    """
    return OrderedDict(
        [
            (k.split(".", 1)[1], v) if k.startswith("_orig_mod") else (k, v)
            for k, v in state_dict.items()
        ]
    )



import copy
from collections import OrderedDict

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from agilerl.networks.custom_components import GumbelSoftmax, NoisyLinear


class MakeEvolvable(nn.Module):
    """Wrapper to make a neural network evolvable

    :param network: Input neural network
    :type network: nn.Module
    :param input_tensor: Example input tensor so forward pass can be made to detect the network architecture
    :type input_tensor: torch.Tensor
    :param num_atoms: Number of atoms for Rainbow DQN, defaults to 51
    :type num_atoms: int, optional
    :param secondary_input_tensor: Second input tensor if network performs forward pass with two tensors, for example, \
        off-policy algorithms that use a critic(s) with environments that have RGB image observations and thus require CNN \
        architecture, defaults to None
    :type secondary_input_tensor: torch.Tensor, optional
    :param min_hidden_layers: Minimum number of hidden layers the fully connected layer will shrink down to, defaults to 1
    :type min_hidden_layers: int, optional
    :param max_hidden_layers: Maximum number of hidden layers the fully connected layer will expand to, defaults to 3
    :type max_hidden_layers: int, optional
    :param min_mlp_nodes: Minimum number of nodes a layer can have within the fully connected layer, defaults to 64
    :type min_mlp_nodes: int, optional
    :param max_mlp_nodes: Maximum number of nodes a layer can have within the fully connected layer, defaults to 1024
    :type max_mlp_nodes: int, optional
    :param min_cnn_hidden_layers: Minimum number of hidden layers the convolutional layer will shrink down to, defaults to 1
    :type min_cnn_hidden_layers: int, optional
    :param max_cnn_hidden_layers: Maximum number of hidden layers the convolutional layer will expand to, defaults to 6
    :type max_cnn_hidden_layers: int, optional
    :param min_channel_size: Minimum number of channels a convolutional layer can have, defaults to 32
    :type min_channel_size: int, optional
    :param max_channel_size: Maximum number of channels a convolutional layer can have, defaults to 256
    :type max_channel_size: int, optional
    :param output_vanish: Vanish output by multiplying by 0.1, defaults to False
    :type output_vanish: bool, optional
    :param init_layers: Initialise network layers, defaults to False
    :type init_layers: bool, optional
    :param support: Atoms support tensor, defaults to None
    :type support: torch.Tensor(), optional
    :param rainbow: Using Rainbow DQN, defaults to False
    :type rainbow: bool, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    """

    def __init__(
        self,
        network,
        input_tensor,
        secondary_input_tensor=None,
        num_atoms=51,
        min_hidden_layers=1,
        max_hidden_layers=3,
        min_mlp_nodes=64,
        max_mlp_nodes=1024,
        min_cnn_hidden_layers=1,
        max_cnn_hidden_layers=6,
        min_channel_size=32,
        max_channel_size=256,
        output_vanish=False,
        init_layers=False,
        support=None,
        rainbow=False,
        device="cpu",
        accelerator=None,
        **kwargs,
    ):
        super().__init__()
        assert (
            min_hidden_layers < max_hidden_layers
        ), "'min_hidden_layers' must be less than 'max_hidden_layers."
        assert (
            min_mlp_nodes < max_mlp_nodes
        ), "'min_mlp_nodes' must be less than 'max_mlp_nodes."
        assert (
            min_cnn_hidden_layers < max_cnn_hidden_layers
        ), "'min_cnn_hidden_layers' must be less than 'max_cnn_hidden_layers."
        assert (
            min_channel_size < max_channel_size
        ), "'min_channel_size' must be less than 'max_channel_size'."
        if not kwargs:
            assert isinstance(
                network, nn.Module
            ), f"'network' must be of type 'nn.Module'.{type(network)}"

        self.init_layers = init_layers
        self.min_hidden_layers = min_hidden_layers
        self.max_hidden_layers = max_hidden_layers
        self.min_mlp_nodes = min_mlp_nodes
        self.max_mlp_nodes = max_mlp_nodes
        self.min_cnn_hidden_layers = min_cnn_hidden_layers
        self.max_cnn_hidden_layers = max_cnn_hidden_layers
        self.min_channel_size = min_channel_size
        self.max_channel_size = max_channel_size
        self.output_vanish = output_vanish
        self.device = device
        self.accelerator = accelerator

        #### Rainbow attributes
        self.rainbow = rainbow  #### add in as a doc string
        self.num_atoms = num_atoms
        self.support = support

        # Set the layer counters
        self.conv_counter = -1
        self.lin_counter = -1
        self.extra_critic_dims = (
            secondary_input_tensor.shape[-1]
            if secondary_input_tensor is not None
            else None
        )

        # Set placeholder attributes (needed for init_dict function to work)
        self.has_conv_layers = False
        self.input_tensor = input_tensor.to(self.device)
        self.secondary_input_tensor = (
            secondary_input_tensor.to(self.device)
            if secondary_input_tensor is not None
            else secondary_input_tensor
        )
        (
            self.in_channels,
            self.channel_size,
            self.kernel_size,
            self.stride_size,
            self.padding,
        ) = (None, None, None, None, None)

        # If first instance, network used to instantiate, upon cloning, init_dict used instead
        if not kwargs:
            self.detect_architecture(
                network.to(self.device), self.input_tensor, self.secondary_input_tensor
            )
        else:
            for key, value in kwargs.items():
                setattr(self, key, value)

        self.feature_net, self.value_net, self.advantage_net = self.create_nets()

    def forward(self, x, xc=None, q=True):
        """Returns output of neural network.

        :param x: Neural network input
        :type x: torch.Tensor() or np.array
        :param xc: Actions to be evaluated by critic, defaults to None
        :type xc: torch.Tensor() or np.array, optional
        :param q: Return Q value if using rainbow, defaults to True
        :type q: bool, optional
        """
        if not isinstance(x, torch.Tensor):
            x = torch.FloatTensor(np.array(x))

        if self.accelerator is None:
            x = x.to(self.device)

        if x.dtype != torch.float32:
            x = x.type(torch.float32)

        batch_size = x.size(0)

        x = self.feature_net(x)

        # Check if there is a cnn
        if self.cnn_layer_info:
            x = x.reshape(batch_size, -1)
            # Ensure dtype is float32

            # Concatenate actions if passed to network as a separate tensor
            if xc is not None:
                if self.accelerator is None:
                    xc = xc.to(self.device)
                x = torch.cat([x, xc], dim=1)

            value = self.value_net(x)

        # add in cnn functionality
        if self.rainbow:
            advantage = self.advantage_net(x)
            if not self.cnn_layer_info:
                value = self.value_net(x)
                value = value.view(-1, 1, self.num_atoms)
                advantage = advantage.view(-1, self.num_outputs, self.num_atoms)
                x = value + advantage - advantage.mean(1, keepdim=True)
                x = F.softmax(x, dim=-1)
            else:
                value = value.view(batch_size, 1, self.num_atoms)
                advantage = advantage.view(batch_size, self.num_outputs, self.num_atoms)

                x = value + advantage - advantage.mean(1, keepdim=True)
                x = F.softmax(x.view(-1, self.num_atoms), dim=-1).view(
                    -1, self.num_outputs, self.num_atoms
                )
            x = x.clamp(min=1e-3)

            if q:
                x = torch.sum(x * self.support, dim=2)
        else:
            if self.cnn_layer_info:
                x = value

        return x

    def layer_init(self, layer, std=np.sqrt(2), bias_const=0.0):
        """Initialize the weights of a neural network layer using orthogonal initialization and set the biases to a constant value.

        :param layer: Neural network layer
        :type layer: nn.Module
        :param std: Standard deviation, defaults to sqrt(2)
        :type std: float
        :param bias_const: Bias value, defaults to 0.0
        :type bias_const: float
        """
        torch.nn.init.orthogonal_(layer.weight, std)
        torch.nn.init.constant_(layer.bias, bias_const)
        return layer

    def get_pooling(self, pooling_names, kernel_size, stride, padding):
        """Returns pooling layer for corresponding activation name.

        :param pooling_names: Pooling layer name
        :type pooling_names: str
        :param kernel_size: Pooling layer kernel size
        :type kernel_size: int or Tuple[int]
        :param stride: Pooling layer stride
        :type stride: int or Tuple[int]
        :param padding: Pooling layer padding
        :type padding: int or Tuple[int]
        """
        pooling_functions = {
            "MaxPool2d": nn.MaxPool2d,
            "MaxPool3d": nn.MaxPool3d,
            "AvgPool2d": nn.AvgPool2d,
            "AvgPool3d": nn.AvgPool3d,
        }

        return pooling_functions[pooling_names](kernel_size, stride, padding)

    def get_activation(self, activation_names):
        """Returns activation function for corresponding activation name.

        :param activation_names: Activation function name
        :type activation_names: str
        """
        activation_functions = {
            "Tanh": nn.Tanh,
            "Linear": nn.Identity,
            "ReLU": nn.ReLU,
            "ELU": nn.ELU,
            "Softsign": nn.Softsign,
            "Sigmoid": nn.Sigmoid,
            "GumbelSoftmax": GumbelSoftmax,
            "Softplus": nn.Softplus,
            "Softmax": nn.Softmax,
            "LeakyReLU": nn.LeakyReLU,
            "PReLU": nn.PReLU,
            "GELU": nn.GELU,
        }

        return (
            activation_functions[activation_names](dim=-1)
            if activation_names == "Softmax"
            else activation_functions[activation_names]()
        )

    def get_normalization(self, normalization_name, layer_size):
        """Returns normalization layer for corresponding normalization name.

        :param normalization_names: Normalization layer name
        :type normalization_names: str
        :param layer_size: The layer after which the normalization layer will be applied
        :param layer_size: int
        """
        normalization_functions = {
            "BatchNorm2d": nn.BatchNorm2d,
            "BatchNorm3d": nn.BatchNorm3d,
            "InstanceNorm2d": nn.InstanceNorm2d,
            "InstanceNorm3d": nn.InstanceNorm3d,
            "LayerNorm": nn.LayerNorm,
        }

        return normalization_functions[normalization_name](layer_size)

    def get_conv_layer(
        self, conv_layer_name, in_channels, out_channels, kernel_size, stride, padding
    ):
        """Return convolutional layer for corresponding convolutional layer name.

        :param conv_layer_name: Convolutional layer name
        :type conv_layer_name: str
        :param in_channels: Number of input channels to convolutional layer
        :type in_channels: int
        :param out_channels: Number of output channels from convolutional layer
        :type out_channels: int
        :param kernel_size: Kernel size of convolutional layer
        :type kernel_size: int or Tuple[int]
        :param stride: Stride size of convolutional layer
        :type stride: int or Tuple[int]
        :param padding: Convolutional layer padding
        :type padding: int or Tuple[int]
        """

        convolutional_layers = {"Conv2d": nn.Conv2d, "Conv3d": nn.Conv3d}

        return convolutional_layers[conv_layer_name](
            in_channels, out_channels, kernel_size, stride, padding
        )

    def detect_architecture(self, network, input_tensor, secondary_input_tensor=None):
        """Detect the architecture of a neural network.

        :param network: Neural network whose architecture is being detected
        :type network: nn.Module
        :param input_tensor: Tensor used to perform forward pass to detect layers
        :type input_tensor: torch.Tensor
        :param secondary_input_tensor: Second tensor used to perform forward pass if forward
        method of neural network takes two tensors as arguments, defaults to None
        :type secondary_input_tensor: torch.Tensor, optional
        """
        in_features_list = []
        out_features_list = []
        in_channel_list = []
        out_channel_list = []
        kernel_size_list = []
        stride_size_list = []
        padding_list = []

        cnn_layer_info = dict()
        mlp_layer_info = dict()

        def register_hooks(module):
            def forward_hook(module, input, output):
                # Convolutional layer detection
                if isinstance(module, nn.modules.conv._ConvNd):
                    self.has_conv_layers = True
                    self.conv_counter += 1

                    if "conv_layer_type" not in cnn_layer_info.keys():
                        cnn_layer_info["conv_layer_type"] = str(
                            module.__class__.__name__
                        )

                    in_channel_list.append(module.in_channels)
                    out_channel_list.append(module.out_channels)
                    kernel_size_list.append(module.kernel_size)
                    stride_size_list.append(module.stride)
                    padding_list.append(module.padding)

                # Linear layer detection
                elif isinstance(module, nn.Linear):
                    self.lin_counter += 1
                    in_features_list.append(module.in_features)
                    out_features_list.append(module.out_features)

                # Normalization layer detection
                elif isinstance(
                    module,
                    (
                        nn.BatchNorm2d,
                        nn.BatchNorm3d,
                        nn.InstanceNorm2d,
                        nn.InstanceNorm3d,
                        nn.LayerNorm,
                    ),
                ):
                    if len(output.shape) <= 2:
                        if "norm_layers" not in mlp_layer_info.keys():
                            mlp_layer_info["norm_layers"] = dict()
                        mlp_layer_info["norm_layers"][self.lin_counter] = str(
                            module.__class__.__name__
                        )
                    else:
                        if "norm_layers" not in cnn_layer_info.keys():
                            cnn_layer_info["norm_layers"] = dict()
                        cnn_layer_info["norm_layers"][self.conv_counter] = str(
                            module.__class__.__name__
                        )

                # Pooling layer detection
                elif isinstance(
                    module, (nn.MaxPool2d, nn.MaxPool3d, nn.AvgPool2d, nn.AvgPool3d)
                ):
                    if "pooling_layers" not in cnn_layer_info.keys():
                        cnn_layer_info["pooling_layers"] = dict()  #
                    cnn_layer_info["pooling_layers"][self.conv_counter] = dict()  #
                    cnn_layer_info["pooling_layers"][self.conv_counter]["name"] = str(
                        module.__class__.__name__
                    )
                    cnn_layer_info["pooling_layers"][self.conv_counter][
                        "kernel"
                    ] = module.kernel_size
                    cnn_layer_info["pooling_layers"][self.conv_counter][
                        "stride"
                    ] = module.stride
                    cnn_layer_info["pooling_layers"][self.conv_counter][
                        "padding"
                    ] = module.padding

                # Skip nn.Flatten layer as this is added when building the CNN to connect
                # the convolutional layers with the fully-connected layers
                elif isinstance(module, nn.Flatten):
                    pass

                # Detect activation layer (supported currently by AgileRL)
                elif isinstance(
                    module,
                    (
                        nn.Tanh,
                        nn.Identity,
                        nn.ReLU,
                        nn.ELU,
                        nn.Softsign,
                        nn.Sigmoid,
                        GumbelSoftmax,
                        nn.Softplus,
                        nn.Softmax,
                        nn.LeakyReLU,
                        nn.PReLU,
                        nn.GELU,
                    ),
                ):
                    if len(output.shape) <= 2:
                        if "activation_layers" not in mlp_layer_info.keys():
                            mlp_layer_info["activation_layers"] = dict()
                        mlp_layer_info["activation_layers"][self.lin_counter] = str(
                            module.__class__.__name__
                        )
                    else:
                        if "activation_layers" not in cnn_layer_info.keys():
                            cnn_layer_info["activation_layers"] = dict()
                        cnn_layer_info["activation_layers"][self.conv_counter] = str(
                            module.__class__.__name__
                        )
                else:
                    raise Exception(
                        f"{module} not currently supported, use an alternative layer."
                    )

            if (
                not isinstance(module, nn.Sequential)
                and not isinstance(module, nn.ModuleList)
                and not isinstance(module, type(network))
            ):
                hooks.append(module.register_forward_hook(forward_hook))

        hooks = []
        network.apply(register_hooks)

        # Forward pass to collect network data necessary to make network evolvable
        with torch.no_grad():
            if secondary_input_tensor is None:
                network(input_tensor)
            else:
                network(input_tensor, secondary_input_tensor)

        # Remove hooks
        for hook in hooks:
            hook.remove()

        # Save neural network information as attributes
        self.num_inputs, *self.hidden_size = in_features_list
        *_, self.num_outputs = out_features_list
        if len(self.hidden_size) == 0:
            raise TypeError("Network must have at least one hidden layer.")
        self.mlp_layer_info = mlp_layer_info

        if len(out_features_list) - 1 in mlp_layer_info["activation_layers"].keys():
            self.mlp_output_activation = mlp_layer_info["activation_layers"][
                len(out_features_list) - 1
            ]
        else:
            self.mlp_output_activation = None
        activation_function_set = set(mlp_layer_info["activation_layers"].values())
        if self.mlp_output_activation is not None:
            activation_function_set.remove(self.mlp_output_activation)
        if len(activation_function_set) > 1:
            raise TypeError(
                "All activation functions other than the output layer activation must be the same."
            )
        else:
            self.mlp_activation = list(mlp_layer_info["activation_layers"].values())[0]

        self.cnn_layer_info = cnn_layer_info
        if self.has_conv_layers is True:
            self.arch = "cnn"
            self.in_channels = in_channel_list[0]
            self.channel_size = out_channel_list
            self.kernel_size = kernel_size_list
            self.stride_size = stride_size_list
            self.padding = padding_list
        else:
            self.arch = "mlp"

        # Reset the layer counters
        self.conv_counter = -1
        self.lin_counter = -1

    def create_mlp(
        self,
        input_size,
        output_size,
        hidden_size,
        name,
        mlp_activation,
        mlp_output_activation,
        noisy=False,
        rainbow_feature_net=False,
    ):
        """Creates and returns multi-layer perceptron.

        :param input_size: Input dimensions to first MLP layer
        :type input_size: int
        :param output_size: Output dimensions from last MLP layer
        :type output_size: int
        :param hidden_size: Hidden layer sizes
        :type hidden_size: list[int]
        :param name: Layer name
        :type name: str
        ####
        :param noisy:
        :type noisy:
        :param rainbow_feature_net:
        :type rainbow_feature_net:
        """

        net_dict = OrderedDict()
        if noisy:
            net_dict[f"{name}_linear_layer_0"] = NoisyLinear(input_size, hidden_size[0])
        else:
            net_dict[f"{name}_linear_layer_0"] = nn.Linear(input_size, hidden_size[0])

        if self.init_layers:
            net_dict[f"{name}_linear_layer_0"] = self.layer_init(
                net_dict[f"{name}_linear_layer_0"]
            )

        if ("norm_layers" in self.mlp_layer_info.keys()) and (
            0 in self.mlp_layer_info["norm_layers"].keys()
        ):
            net_dict[f"{name}_layer_norm_0"] = self.get_normalization(
                self.mlp_layer_info["norm_layers"][0], hidden_size[0]
            )

        if ("activation_layers" in self.mlp_layer_info.keys()) and (
            0 in self.mlp_layer_info["activation_layers"].keys()
        ):
            net_dict[f"{name}_activation_0"] = self.get_activation(
                mlp_output_activation
                if (len(hidden_size) == 1 and rainbow_feature_net)
                else mlp_activation
            )

        if len(hidden_size) > 1:
            for l_no in range(1, len(hidden_size)):
                if noisy:
                    net_dict[f"{name}_linear_layer_{str(l_no)}"] = NoisyLinear(
                        hidden_size[l_no - 1], hidden_size[l_no]
                    )
                else:
                    net_dict[f"{name}_linear_layer_{str(l_no)}"] = nn.Linear(
                        hidden_size[l_no - 1], hidden_size[l_no]
                    )
                if self.init_layers:
                    net_dict[f"{name}_linear_layer_{str(l_no)}"] = self.layer_init(
                        net_dict[f"{name}_linear_layer_{str(l_no)}"]
                    )
                if ("norm_layers" in self.mlp_layer_info.keys()) and (
                    l_no in self.mlp_layer_info["norm_layers"].keys()
                ):
                    net_dict[f"{name}_layer_norm_{str(l_no)}"] = self.get_normalization(
                        self.mlp_layer_info["norm_layers"][l_no], hidden_size[l_no]
                    )
                if l_no in self.mlp_layer_info["activation_layers"].keys():
                    net_dict[f"{name}_activation_{str(l_no)}"] = self.get_activation(
                        mlp_activation
                        if not rainbow_feature_net
                        else mlp_output_activation
                    )
        if not rainbow_feature_net:
            if noisy:
                output_layer = NoisyLinear(hidden_size[-1], output_size)
            else:
                output_layer = nn.Linear(hidden_size[-1], output_size)
            if self.init_layers:
                output_layer = self.layer_init(output_layer)

            if self.output_vanish:
                output_layer.weight.data.mul_(0.1)
                output_layer.bias.data.mul_(0.1)

            net_dict[f"{name}_linear_layer_output"] = output_layer
            if mlp_output_activation is not None:
                net_dict[f"{name}_activation_output"] = self.get_activation(
                    mlp_output_activation
                )

        return nn.Sequential(net_dict)

    def create_cnn(
        self, input_size, channel_size, kernel_size, stride_size, padding, name
    ):
        """Creates and returns convolutional neural network.

        :param input_size: Channel size of first layer
        :type input_size: int
        :param channel_size: Output channel sizes for each layer
        :type channel_size: list[int]
        :param kernel_size: Kernel sizes
        :type kernel_size: list[int] or list[Tuple[int]]
        :param stride_size: Stride sizes
        :type stride_size: list[int] or list[Tuple[int]]
        :param padding: Convolutional layer padding
        :type padding: list[int] or list[Tuple[int]]
        :param name: Layer name
        :type name: str
        """

        net_dict = OrderedDict()
        # if self.cnn_layer_info["conv_layer_type"] == "Conv3d":
        #     k_size = [
        #         (self.input_tensor.shape[-3], k_size[1], k_size[2])
        #         for k_size in kernel_size
        #     ]
        # else:

        net_dict[f"{name}_conv_layer_0"] = self.get_conv_layer(
            self.cnn_layer_info["conv_layer_type"],
            in_channels=input_size,
            out_channels=channel_size[0],
            kernel_size=kernel_size[0],
            stride=stride_size[0],
            padding=padding[0],
        )
        if ("norm_layers" in self.cnn_layer_info.keys()) and (
            0 in self.cnn_layer_info["norm_layers"].keys()
        ):
            net_dict[f"{name}_layer_norm_0"] = self.get_normalization(
                self.cnn_layer_info["norm_layers"][0], channel_size[0]
            )

        if ("activation_layers" in self.cnn_layer_info.keys()) and (
            0 in self.cnn_layer_info["activation_layers"].keys()
        ):
            net_dict[f"{name}_activation_0"] = self.get_activation(
                self.cnn_layer_info["activation_layers"][0]
            )

        if ("pooling_layers" in self.cnn_layer_info.keys()) and (
            0 in self.cnn_layer_info["pooling_layers"].keys()
        ):
            net_dict[f"{name}_pooling_0"] = self.get_pooling(
                self.cnn_layer_info["pooling_layers"][0]["name"],
                self.cnn_layer_info["pooling_layers"][0]["kernel"],
                self.cnn_layer_info["pooling_layers"][0]["stride"],
                self.cnn_layer_info["pooling_layers"][0]["padding"],
            )

        if len(channel_size) > 1:
            for l_no in range(1, len(channel_size)):
                net_dict[f"{name}_conv_layer_{str(l_no)}"] = self.get_conv_layer(
                    self.cnn_layer_info["conv_layer_type"],
                    in_channels=channel_size[l_no - 1],
                    out_channels=channel_size[l_no],
                    kernel_size=kernel_size[l_no],
                    stride=stride_size[l_no],
                    padding=padding[l_no],
                )
                if ("norm_layers" in self.cnn_layer_info.keys()) and (
                    l_no in self.cnn_layer_info["norm_layers"].keys()
                ):
                    net_dict[f"{name}_layer_norm_{str(l_no)}"] = self.get_normalization(
                        self.cnn_layer_info["norm_layers"][l_no], channel_size[l_no]
                    )

                if ("activation_layers" in self.cnn_layer_info.keys()) and (
                    l_no in self.cnn_layer_info["activation_layers"].keys()
                ):
                    net_dict[f"{name}_activation_{str(l_no)}"] = self.get_activation(
                        self.cnn_layer_info["activation_layers"][l_no]
                    )

                if ("pooling_layers" in self.cnn_layer_info.keys()) and (
                    l_no in self.cnn_layer_info["pooling_layers"].keys()
                ):
                    net_dict[f"{name}_pooling_{str(l_no)}"] = self.get_pooling(
                        self.cnn_layer_info["pooling_layers"][l_no]["name"],
                        self.cnn_layer_info["pooling_layers"][l_no]["kernel"],
                        self.cnn_layer_info["pooling_layers"][l_no]["stride"],
                        self.cnn_layer_info["pooling_layers"][l_no]["padding"],
                    )

        return nn.Sequential(net_dict)

    def create_nets(self):
        """Creates and returns the feature and value net."""

        # Check if any CNN layers otherwise return just a mlp
        if self.cnn_layer_info:
            feature_net = self.create_cnn(
                self.in_channels,
                self.channel_size,
                self.kernel_size,
                self.stride_size,
                self.padding,
                name="feature",
            )
            cnn_output = feature_net(torch.zeros(*self.input_tensor.shape))
            self.cnn_output_size = cnn_output.shape
            input_size = (cnn_output).to(self.device).view(1, -1).size(1)

            if self.secondary_input_tensor is not None:
                input_size += self.extra_critic_dims

            if self.rainbow:
                value_net = self.create_mlp(
                    input_size,
                    output_size=self.num_atoms,
                    hidden_size=self.hidden_size,
                    name="value",
                    noisy=True,
                    mlp_output_activation=self.mlp_output_activation,
                    mlp_activation=self.mlp_activation,
                )
                advantage_net = self.create_mlp(
                    input_size,
                    output_size=self.num_atoms * self.num_outputs,  ####
                    hidden_size=self.hidden_size,
                    name="advantage",
                    noisy=True,
                    mlp_output_activation=self.mlp_output_activation,
                    mlp_activation=self.mlp_activation,
                )
            else:
                value_net = self.create_mlp(
                    input_size,
                    self.num_outputs,
                    self.hidden_size,
                    name="value",
                    mlp_activation=self.mlp_activation,
                    mlp_output_activation=self.mlp_output_activation,
                )
                advantage_net = None

        else:
            input_size = self.num_inputs
            if self.rainbow:
                feature_net = self.create_mlp(
                    input_size=self.num_inputs,
                    output_size=128,
                    hidden_size=[128],
                    name="feature",
                    rainbow_feature_net=True,
                    mlp_activation=self.mlp_activation,
                    mlp_output_activation="ReLU",
                )
                value_net = self.create_mlp(
                    input_size=128,
                    output_size=self.num_atoms,
                    hidden_size=self.hidden_size,
                    noisy=True,
                    name="value",
                    mlp_output_activation=self.mlp_output_activation,
                    mlp_activation=self.mlp_activation,
                )
                advantage_net = self.create_mlp(
                    input_size=128,
                    output_size=self.num_atoms * self.num_outputs,
                    hidden_size=self.hidden_size,
                    noisy=True,
                    name="advantage",
                    mlp_output_activation=self.mlp_output_activation,
                    mlp_activation=self.mlp_activation,
                )
            else:
                value_net = None
                advantage_net = None
                feature_net = self.create_mlp(
                    input_size,
                    self.num_outputs,
                    self.hidden_size,
                    name="feature",
                    mlp_activation=self.mlp_activation,
                    mlp_output_activation=self.mlp_output_activation,
                )

        if self.accelerator is None:
            feature_net = feature_net.to(self.device)
            value_net = (
                value_net.to(self.device) if value_net is not None else value_net
            )
            advantage_net = (
                advantage_net.to(self.device)
                if advantage_net is not None
                else advantage_net
            )
        return feature_net, value_net, advantage_net

    @property
    def init_dict(self):
        """Returns model information in dictionary."""
        init_dict = {
            "network": None,
            "input_tensor": self.input_tensor,
            "secondary_input_tensor": self.secondary_input_tensor,
            "num_inputs": self.num_inputs,
            "num_outputs": self.num_outputs,
            "hidden_size": self.hidden_size,
            "mlp_activation": self.mlp_activation,
            "mlp_output_activation": self.mlp_output_activation,
            "device": self.device,
            "accelerator": self.accelerator,
            "in_channels": self.in_channels,
            "channel_size": self.channel_size,
            "kernel_size": self.kernel_size,
            "stride_size": self.stride_size,
            "padding": self.padding,
            "extra_critic_dims": self.extra_critic_dims,
            "output_vanish": self.output_vanish,
            "init_layers": self.init_layers,
            "has_conv_layer": self.has_conv_layers,
            "arch": self.arch,
            "cnn_layer_info": self.cnn_layer_info,
            "mlp_layer_info": self.mlp_layer_info,
            "num_atoms": self.num_atoms,
            "rainbow": self.rainbow,
            "support": self.support,
        }

        return init_dict

    def add_mlp_layer(self):
        """Adds a hidden layer to value network."""
        if len(self.hidden_size) < self.max_hidden_layers:  # HARD LIMIT
            self.hidden_size += [self.hidden_size[-1]]
            self.mlp_layer_info["activation_layers"][
                len(self.hidden_size) - 1
            ] = self.mlp_activation
            if self.mlp_output_activation is not None:
                self.mlp_layer_info["activation_layers"][
                    len(self.hidden_size)
                ] = self.mlp_output_activation
            self.recreate_nets()
        else:
            self.add_mlp_node()

    def remove_mlp_layer(self):
        """Removes a hidden layer from value network."""
        if len(self.hidden_size) > self.min_hidden_layers:  # HARD LIMIT
            self.hidden_size = self.hidden_size[:-1]
            if len(self.hidden_size) in self.mlp_layer_info["activation_layers"].keys():
                if self.mlp_output_activation is None:
                    self.mlp_layer_info["activation_layers"].pop(len(self.hidden_size))
                else:
                    self.mlp_layer_info["activation_layers"].pop(
                        len(self.hidden_size) + 1
                    )
                    self.mlp_layer_info["activation_layers"][
                        len(self.hidden_size)
                    ] = self.mlp_output_activation
            else:
                if self.mlp_output_activation is not None:
                    self.mlp_layer_info["activation_layers"].pop(
                        len(self.hidden_size) + 1
                    )
                    self.mlp_layer_info["activation_layers"][
                        len(self.hidden_size)
                    ] = self.mlp_output_activation

            if (
                "norm_layers" in self.mlp_layer_info.keys()
                and len(self.hidden_size) in self.mlp_layer_info["norm_layers"]
            ):
                self.mlp_layer_info["norm_layers"].pop(len(self.hidden_size))

            self.recreate_nets(shrink_params=True)
        else:
            self.add_mlp_node()

    def add_mlp_node(self, hidden_layer=None, numb_new_nodes=None):
        """Adds nodes to hidden layer of value network.

        :param hidden_layer: Depth of hidden layer to add nodes to, defaults to None
        :type hidden_layer: int, optional
        :param numb_new_nodes: Number of nodes to add to hidden layer, defaults to None
        :type numb_new_nodes: int, optional
        """
        if hidden_layer is None:
            hidden_layer = np.random.randint(0, len(self.hidden_size), 1)[0]
        else:
            hidden_layer = min(hidden_layer, len(self.hidden_size) - 1)
        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([16, 32, 64], 1)[0]

        if (
            self.hidden_size[hidden_layer] + numb_new_nodes <= self.max_mlp_nodes
        ):  # HARD LIMIT
            self.hidden_size[hidden_layer] += numb_new_nodes

            self.recreate_nets()

        return {"hidden_layer": hidden_layer, "numb_new_nodes": numb_new_nodes}

    def remove_mlp_node(self, hidden_layer=None, numb_new_nodes=None):
        """Removes nodes from hidden layer of neural network.

        :param hidden_layer: Depth of hidden layer to remove nodes from, defaults to None
        :type hidden_layer: int, optional
        :param numb_new_nodes: Number of nodes to remove from hidden layer, defaults to None
        :type numb_new_nodes: int, optional
        """
        if hidden_layer is None:
            hidden_layer = np.random.randint(0, len(self.hidden_size), 1)[0]
        else:
            hidden_layer = min(hidden_layer, len(self.hidden_size) - 1)
        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([16, 32, 64], 1)[0]

        if (
            self.hidden_size[hidden_layer] - numb_new_nodes > self.min_mlp_nodes
        ):  # HARD LIMIT
            self.hidden_size[hidden_layer] -= numb_new_nodes
            self.recreate_nets(shrink_params=True)

        return {"hidden_layer": hidden_layer, "numb_new_nodes": numb_new_nodes}

    def add_cnn_layer(self):
        """Adds a hidden layer to convolutional neural network."""
        max_kernels = self.calc_max_kernel_sizes()
        stride_size_ranges = self.calc_stride_size_ranges()
        if (
            len(self.channel_size) < self.max_cnn_hidden_layers
            and not any(i <= 2 for i in self.cnn_output_size[-2:])
            and all(i > 0 for i in max_kernels[-1])
        ):  # HARD LIMIT
            self.channel_size += [self.channel_size[-1]]
            k_size = tuple(np.random.randint(1, 1 + k) for k in max_kernels[-1])
            self.kernel_size += [k_size]
            self.padding += [self.padding[-1]]
            stride_size_list = [
                tuple(
                    np.random.randint(tup[0], tup[1] + 1) for _ in self.stride_size[-1]
                )
                for tup in stride_size_ranges
            ]
            self.stride_size = stride_size_list + [
                tuple(1 for _ in self.stride_size[-1])
            ]
            if "activation_layers" not in self.cnn_layer_info.keys():
                self.cnn_layer_info["activation_layers"] = dict()
            self.cnn_layer_info["activation_layers"][
                len(self.channel_size) - 1
            ] = "ReLU"

            self.recreate_nets()
        else:
            self.add_cnn_channel()

    def remove_cnn_layer(self):
        """Removes a hidden layer from the convolutional neural network."""
        stride_size_ranges = self.calc_stride_size_ranges()
        if len(self.channel_size) > self.min_cnn_hidden_layers:
            self.channel_size = self.channel_size[:-1]
            self.kernel_size = self.kernel_size[:-1]

            stride_size_list = [
                tuple(
                    np.random.randint(tup[0], tup[1] + 1) for _ in self.stride_size[-1]
                )
                for tup in stride_size_ranges
            ]
            self.stride_size = stride_size_list[:-1]

            if "activation_layers" in self.cnn_layer_info.keys():
                if len(self.channel_size) in self.cnn_layer_info["activation_layers"]:
                    self.cnn_layer_info["activation_layers"].pop(len(self.channel_size))
            else:
                self.cnn_layer_info["activation_layers"] = dict()
            if (
                len(self.channel_size) - 1
                not in self.cnn_layer_info["activation_layers"]
            ):
                self.cnn_layer_info["activation_layers"][
                    len(self.channel_size) - 1
                ] = "ReLU"

            if (
                "pooling_layers" in self.cnn_layer_info.keys()
                and len(self.channel_size) in self.cnn_layer_info["pooling_layers"]
            ):
                self.cnn_layer_info["pooling_layers"].pop(len(self.channel_size))

            if (
                "norm_layers" in self.cnn_layer_info.keys()
                and len(self.channel_size) in self.cnn_layer_info["norm_layers"]
            ):
                self.cnn_layer_info["norm_layers"].pop(len(self.channel_size))

            self.recreate_nets()
        else:
            self.add_cnn_channel()

    def calc_max_kernel_sizes(self):
        "Calculates the max kernel size for each convolutional layer of the feature net."
        max_kernel_list = []
        if self.cnn_layer_info["conv_layer_type"] != "Conv3d":
            height_in, width_in = self.input_tensor.shape[-2:]
            for idx, _ in enumerate(self.channel_size):
                height_out = 1 + (
                    height_in
                    + 2 * self.padding[idx][0]
                    - 1 * (self.kernel_size[idx][0] - 1)
                    - 1
                ) / (self.stride_size[idx][0])
                width_out = 1 + (
                    width_in
                    + 2 * self.padding[idx][0]
                    - 1 * (self.kernel_size[idx][1] - 1)
                    - 1
                ) / (self.stride_size[idx][1])
                max_kernel_sizes = np.array([height_out * 0.2, width_out * 0.2])
                max_kernel_sizes = np.where(max_kernel_sizes < 0, 0, max_kernel_sizes)
                max_kernel_sizes = np.where(max_kernel_sizes > 10, 10, max_kernel_sizes)
                max_kernel_list.append(tuple(max_kernel_sizes))
                height_in = height_out
                width_in = width_out
        else:
            depth_in, height_in, width_in = self.input_tensor.shape[-3:]
            for idx, _ in enumerate(self.channel_size):
                depth_out = 1 + (
                    depth_in
                    + 2 * self.padding[idx][0]
                    - 1 * (self.kernel_size[idx][0] - 1)
                    - 1
                ) / (self.stride_size[idx][0])
                height_out = 1 + (
                    height_in
                    + 2 * self.padding[idx][1]
                    - 1 * (self.kernel_size[idx][1] - 1)
                    - 1
                ) / (self.stride_size[idx][1])
                width_out = 1 + (
                    width_in
                    + 2 * self.padding[idx][2]
                    - 1 * (self.kernel_size[idx][2] - 1)
                    - 1
                ) / (self.stride_size[idx][2])
                max_kernel_sizes = np.array(
                    [depth_out, height_out * 0.2, width_out * 0.2]
                )
                max_kernel_sizes = np.where(max_kernel_sizes < 0, 0, max_kernel_sizes)
                max_kernel_sizes = np.where(max_kernel_sizes > 10, 10, max_kernel_sizes)
                max_kernel_list.append(tuple(max_kernel_sizes))
                height_in = height_out
                width_in = width_out
                depth_in = depth_out

        return max_kernel_list

    def calc_stride_size_ranges(self):
        "Calculates a range of stride sizes for each convolutional layer of the feature net."
        stride_range_list = []
        if self.cnn_layer_info["conv_layer_type"] != "Conv3d":
            height_in, width_in = self.input_tensor.shape[-2:]
            for idx, _ in enumerate(self.channel_size):
                height_out = 1 + (
                    height_in
                    + 2 * self.padding[idx][0]
                    - 1 * (self.kernel_size[idx][0] - 1)
                    - 1
                ) / (self.stride_size[idx][0])
                width_out = 1 + (
                    width_in
                    + 2 * self.padding[idx][0]
                    - 1 * (self.kernel_size[idx][1] - 1)
                    - 1
                ) / (self.stride_size[idx][1])

                min_stride = min(-(-height_out // 200), -(-width_out // 200))
                max_stride = min(-(-height_out // 75), -(-width_out // 75))

                stride_range_list.append((int(min_stride), int(max_stride)))
                height_in = height_out
                width_in = width_out
        else:
            depth_in, height_in, width_in = self.input_tensor.shape[-3:]
            for idx, _ in enumerate(self.channel_size):
                depth_out = 1 + (
                    depth_in
                    + 2 * self.padding[idx][0]
                    - 1 * (self.kernel_size[idx][0] - 1)
                    - 1
                ) / (self.stride_size[idx][0])
                height_out = 1 + (
                    height_in
                    + 2 * self.padding[idx][1]
                    - 1 * (self.kernel_size[idx][1] - 1)
                    - 1
                ) / (self.stride_size[idx][1])
                width_out = 1 + (
                    width_in
                    + 2 * self.padding[idx][2]
                    - 1 * (self.kernel_size[idx][2] - 1)
                    - 1
                ) / (self.stride_size[idx][2])

                min_stride = min(
                    -(-height_out // 100), -(-width_out // 100), -(-depth_out // 100)
                )
                max_stride = min(
                    -(-height_out // 40), -(-width_out // 40), -(-depth_out // 40)
                )

                stride_range_list.append((int(min_stride), int(max_stride)))
                height_in = height_out
                width_in = width_out
                depth_in = depth_out

        return stride_range_list

    def change_cnn_kernel(self):
        """Randomly alters convolution kernel of random CNN layer."""
        max_kernels = self.calc_max_kernel_sizes()
        # if self.cnn_layer_info["conv_layer_type"] == "Conv3d":
        if len(self.channel_size) > 1:
            hidden_layer = np.random.randint(1, min(4, len(self.channel_size)), 1)[0]
            kernel_size_values = tuple(
                np.random.choice([3, 4, 5, 7]) for _ in self.kernel_size[-1]
            )
            kernel_size_values = tuple(
                i if i <= max_kernels[-1][idx] else int(max_kernels[-1][idx])
                for idx, i in enumerate(kernel_size_values)
            )
            self.kernel_size[hidden_layer] = kernel_size_values
            self.recreate_nets()
        else:
            self.add_cnn_layer()

    def add_cnn_channel(self, hidden_layer=None, numb_new_channels=None):
        """Adds channel to hidden layer of Convolutional Neural Network.

        :param hidden_layer: Depth of hidden layer to add channel to, defaults to None
        :type hidden_layer: int, optional
        :param numb_new_channels: Number of channels to add to hidden layer, defaults to None
        :type numb_new_channels: int, optional
        """
        if hidden_layer is None:
            hidden_layer = np.random.randint(0, len(self.channel_size), 1)[0]
        else:
            hidden_layer = min(hidden_layer, len(self.channel_size) - 1)
        if numb_new_channels is None:
            numb_new_channels = np.random.choice([8, 16, 32], 1)[0]

        if (
            self.channel_size[hidden_layer] + numb_new_channels <= self.max_channel_size
        ):  # HARD LIMIT
            self.channel_size[hidden_layer] += numb_new_channels

            self.recreate_nets()

        return {"hidden_layer": hidden_layer, "numb_new_channels": numb_new_channels}

    def remove_cnn_channel(self, hidden_layer=None, numb_new_channels=None):
        """Remove channel from hidden layer of convolutional neural network.

        :param hidden_layer: Depth of hidden layer to add channel to, defaults to None
        :type hidden_layer: int, optional
        :param numb_new_channels: Number of channels to add to hidden layer, defaults to None
        :type numb_new_channels: int, optional
        """
        if hidden_layer is None:
            hidden_layer = np.random.randint(0, len(self.channel_size), 1)[0]
        else:
            hidden_layer = min(hidden_layer, len(self.channel_size) - 1)
        if numb_new_channels is None:
            numb_new_channels = np.random.choice([8, 16, 32], 1)[0]

        if (
            self.channel_size[hidden_layer] - numb_new_channels > self.min_channel_size
        ):  # HARD LIMIT
            self.channel_size[hidden_layer] -= numb_new_channels

            self.recreate_nets(shrink_params=True)

        return {"hidden_layer": hidden_layer, "numb_new_channels": numb_new_channels}

    def reset_noise(self):
        """Resets noise of value and advantage networks."""
        if self.rainbow:
            for layer in self.value_net:
                if isinstance(layer, NoisyLinear):
                    layer.reset_noise()
            for layer in self.advantage_net:
                if isinstance(layer, NoisyLinear):
                    layer.reset_noise()

    def recreate_nets(self, shrink_params=False):
        """Recreates neural networks.

        :param shrink_params: Boolean flag to shrink parameters
        :type shrink_params: bool
        """
        new_feature_net, new_value_net, new_advantage_net = self.create_nets()

        if shrink_params:
            if self.value_net is not None:
                new_value_net = self.shrink_preserve_parameters(
                    old_net=self.value_net, new_net=new_value_net
                )
            if self.advantage_net is not None:
                new_advantage_net = self.shrink_preserve_parameters(
                    old_net=self.advantage_net, new_net=new_advantage_net
                )
            new_feature_net = self.shrink_preserve_parameters(
                old_net=self.feature_net, new_net=new_feature_net
            )
        else:
            if self.value_net is not None:
                new_value_net = self.preserve_parameters(
                    old_net=self.value_net, new_net=new_value_net
                )
            if self.advantage_net is not None:
                new_advantage_net = self.preserve_parameters(
                    old_net=self.advantage_net, new_net=new_advantage_net
                )
            new_feature_net = self.preserve_parameters(
                old_net=self.feature_net, new_net=new_feature_net
            )

        self.feature_net, self.value_net, self.advantage_net = (
            new_feature_net,
            new_value_net,
            new_advantage_net,
        )

    def clone(self):
        """Returns clone of neural net with identical parameters."""
        clone = MakeEvolvable(**copy.deepcopy(self.init_dict))
        clone.load_state_dict(self.state_dict())
        return clone

    def preserve_parameters(self, old_net, new_net):
        """Returns new neural network with copied parameters from old network.

        :param old_net: Old neural network
        :type old_net: nn.Module()
        :param new_net: New neural network
        :type new_net: nn.Module()
        """
        old_net_dict = dict(old_net.named_parameters())

        for key, param in new_net.named_parameters():
            if key in old_net_dict.keys():
                if old_net_dict[key].data.size() == param.data.size():
                    param.data = old_net_dict[key].data
                else:
                    if "norm" not in key:
                        old_size = old_net_dict[key].data.size()
                        new_size = param.data.size()
                        if len(param.data.size()) == 1:
                            param.data[: min(old_size[0], new_size[0])] = old_net_dict[
                                key
                            ].data[: min(old_size[0], new_size[0])]
                        elif len(param.data.size()) == 2:
                            param.data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                            ] = old_net_dict[key].data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                            ]
                        elif len(param.data.size()) == 3:
                            param.data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                                : min(old_size[2], new_size[2]),
                            ] = old_net_dict[key].data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                                : min(old_size[2], new_size[2]),
                            ]
                        elif len(param.data.size()) == 4:
                            param.data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                                : min(old_size[2], new_size[2]),
                                : min(old_size[3], new_size[3]),
                            ] = old_net_dict[key].data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                                : min(old_size[2], new_size[2]),
                                : min(old_size[3], new_size[3]),
                            ]
                        elif len(param.data.size()) == 5:
                            param.data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                                : min(old_size[2], new_size[2]),
                                : min(old_size[3], new_size[3]),
                                : min(old_size[4], new_size[4]),
                            ] = old_net_dict[key].data[
                                : min(old_size[0], new_size[0]),
                                : min(old_size[1], new_size[1]),
                                : min(old_size[2], new_size[2]),
                                : min(old_size[3], new_size[3]),
                                : min(old_size[4], new_size[4]),
                            ]

        return new_net

    def shrink_preserve_parameters(self, old_net, new_net):
        """Returns shrunk new neural network with copied parameters from old network.

        :param old_net: Old neural network
        :type old_net: nn.Module()
        :param new_net: New neural network
        :type new_net: nn.Module()
        """
        old_net_dict = dict(old_net.named_parameters())

        for key, param in new_net.named_parameters():
            if key in old_net_dict.keys():
                if old_net_dict[key].data.size() == param.data.size():
                    param.data = old_net_dict[key].data
                else:
                    if "norm" not in key:
                        old_size = old_net_dict[key].data.size()
                        new_size = param.data.size()
                        min_0 = min(old_size[0], new_size[0])
                        if len(param.data.size()) == 1:
                            param.data[:min_0] = old_net_dict[key].data[:min_0]
                        else:
                            min_1 = min(old_size[1], new_size[1])
                            param.data[:min_0, :min_1] = old_net_dict[key].data[
                                :min_0, :min_1
                            ]
        return new_net
